import os
import sys

try:
    from unittest import mock, TestCase
except ImportError:
    import mock
from datetime import datetime

from crmsh import utils as crmshutils
from crmsh.crash_test import utils, main, config, task


class TestTaskKill(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    @mock.patch('crmsh.crash_test.utils.msg_info')
    def setUp(self, mock_msg_info):
        """
        Test setUp.
        """
        ctx = mock.Mock(current_case="sbd", loop=False)
        self.task_kill_inst = task.TaskKill(ctx)
        ctx2 = mock.Mock(current_case="sbd", loop=True)
        self.task_kill_inst_loop = task.TaskKill(ctx2)

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    @mock.patch('os.path.isdir')
    def test_enable_report_error(self, mock_isdir):
        mock_isdir.return_value = False
        main.ctx = mock.Mock(report_path="/path")
        with self.assertRaises(task.TaskError) as error:
            self.task_kill_inst.enable_report()
        self.assertEqual("/path is not a directory", str(error.exception))
        mock_isdir.assert_called_once_with("/path")

    @mock.patch('crmsh.crash_test.utils.this_node')
    @mock.patch('crmsh.crash_test.utils.now')
    @mock.patch('os.path.isdir')
    def test_enable_report_looping(self, mock_isdir, mock_now, mock_this_node):
        main.ctx = mock.Mock(report_path="/path", process_name="cpc")
        mock_now.return_value = "20210119-12345"
        mock_this_node.return_value = "node1"
        self.task_kill_inst_loop.enable_report()

    @mock.patch('crmsh.crash_test.utils.this_node')
    @mock.patch('crmsh.crash_test.utils.now')
    @mock.patch('os.path.isdir')
    def test_enable_report(self, mock_isdir, mock_now, mock_this_node):
        main.ctx = mock.Mock(report_path="/path", process_name="cpc")
        mock_now.return_value = "20210119-12345"
        mock_this_node.return_value = "node1"
        self.task_kill_inst.enable_report()

    def test_header(self):
        expected_res = """==============================================
Testcase:          Force kill sbd
Looping Kill:      False
Expected State:    a) sbd process restarted
                   b) Or, this node fenced.
"""
        res = self.task_kill_inst.header()
        self.assertEqual(res, expected_res)

    @mock.patch('crmsh.crash_test.utils.json_dumps')
    def test_to_json(self, mock_dumps):
        self.task_kill_inst.build_base_result = mock.Mock()
        self.task_kill_inst.result = {}
        self.task_kill_inst.prev_task_list = []
        self.task_kill_inst.to_json()
        self.task_kill_inst.build_base_result.assert_called_once_with()
        mock_dumps.assert_called_once_with()

    def test_to_report_return(self):
        self.task_kill_inst.report = False
        self.task_kill_inst.to_report()

    @mock.patch('os.fsync')
    @mock.patch('builtins.open', create=True)
    @mock.patch('crmsh.crash_test.task.TaskKill.header')
    def test_to_report(self, mock_header, mock_open_file, mock_fsync):
        mock_header.return_value = "#### header"
        self.task_kill_inst.report = True
        self.task_kill_inst.messages = [["info", "data", "2021"]]
        self.task_kill_inst.explain = "explain"
        self.task_kill_inst.report_file = "report_file1"
        file_handle = mock_open_file.return_value.__enter__.return_value

        self.task_kill_inst.to_report()

        file_handle.write.assert_has_calls([
            mock.call("#### header"),
            mock.call("\nLog:\n"),
            mock.call("2021 INFO:data\n"),
            mock.call("\nTestcase Explained:\n"),
            mock.call("explain\n")
            ])

    @mock.patch('crmsh.crash_test.utils.get_process_status')
    @mock.patch('crmsh.crash_test.task.Task.task_pre_check')
    def test_pre_check(self, mock_pre_check, mock_status):
        mock_status.return_value = (False, 100)
        with self.assertRaises(task.TaskError) as err:
            self.task_kill_inst.pre_check()
        self.assertEqual("Process sbd is not running!", str(err.exception))
        mock_pre_check.assert_called_once_with()
        mock_status.assert_called_once_with("sbd")

    @mock.patch('crmsh.crash_test.task.TaskKill.process_monitor')
    @mock.patch('crmsh.crash_test.task.Task.fence_action_monitor')
    @mock.patch('threading.Thread')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.Task.info')
    @mock.patch('crmsh.crash_test.utils.get_process_status')
    def test_run(self, mock_status, mock_info, mock_run, mock_thread, mock_fence_monitor, mock_process_monitor):
        mock_status.side_effect = [(False, -1), (True, 100)]
        mock_thread_fence_inst = mock.Mock()
        mock_thread_restart_inst = mock.Mock()
        mock_thread.side_effect = [mock_thread_fence_inst, mock_thread_restart_inst]

        self.task_kill_inst.run()

        mock_status.assert_has_calls([mock.call("sbd"), mock.call("sbd")])
        mock_info.assert_has_calls([
            mock.call('Process sbd(100) is running...'),
            mock.call('Trying to run "killall -9 sbd"')
            ])
        mock_run.assert_called_once_with("killall -9 sbd")
        mock_thread.assert_has_calls([
            mock.call(target=mock_fence_monitor),
            mock.call(target=mock_process_monitor),
            ])
        mock_thread_fence_inst.start.assert_called_once_with()
        mock_thread_restart_inst.start.assert_called_once_with()

    def test_wait_exception(self):
        self.task_kill_inst.fence_start_event = mock.Mock()
        self.task_kill_inst.restart_happen_event = mock.Mock()
        self.task_kill_inst.thread_stop_event = mock.Mock()
        self.task_kill_inst.fence_start_event.wait.return_value = True
        self.task_kill_inst.restart_happen_event.is_set.return_value = False

        with self.assertRaises(task.TaskError) as err:
            self.task_kill_inst.wait()
        self.assertEqual("Process sbd is not restarted!", str(err.exception))

    def test_wait(self):
        self.task_kill_inst.fence_start_event = mock.Mock()
        self.task_kill_inst.restart_happen_event = mock.Mock()
        self.task_kill_inst.thread_stop_event = mock.Mock()
        self.task_kill_inst.fence_start_event.wait.return_value = True
        self.task_kill_inst.restart_happen_event.is_set.return_value = True

        self.task_kill_inst.wait()

        self.task_kill_inst.thread_stop_event.set.assert_called_once_with()

    @mock.patch('time.sleep')
    @mock.patch('crmsh.crash_test.task.Task.info')
    @mock.patch('crmsh.crash_test.utils.get_process_status')
    def test_process_monitor(self, mock_status, mock_info, mock_sleep):
        self.task_kill_inst.thread_stop_event = mock.Mock()
        self.task_kill_inst.thread_stop_event.is_set.side_effect = [False, False]
        self.task_kill_inst.restart_happen_event = mock.Mock()
        mock_status.side_effect = [(False, -1), (True, 100)]

        self.task_kill_inst.process_monitor()

        self.task_kill_inst.thread_stop_event.is_set.assert_has_calls([
            mock.call(),
            mock.call()
            ])
        mock_status.assert_has_calls([
            mock.call("sbd"),
            mock.call("sbd")
            ])
        mock_info.assert_called_once_with("Process sbd(100) is restarted!")
        self.task_kill_inst.restart_happen_event.set.assert_called_once_with()
        mock_sleep.assert_called_once_with(1)


class TestTaskCheck(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    @mock.patch('crmsh.crash_test.utils.msg_info')
    def setUp(self, mock_msg_info):
        """
        Test setUp.
        """
        fake_timedate = datetime(2019, 7, 10, 1, 15, 15)
        main.ctx = mock.Mock(task_list=[{"process_name": "xin", "age": 38}])
        self.task_check_inst = task.TaskCheck("task check job1", quiet=False)
        self.task_check_inst.timestamp = fake_timedate
        self.task_check_inst_quiet = task.TaskCheck("task check job1", quiet=True)

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    @mock.patch('crmsh.crash_test.utils.MyLoggingFormatter')
    @mock.patch('crmsh.crash_test.utils.get_handler')
    @mock.patch('crmsh.crash_test.utils.manage_handler')
    def test_to_stdout(self, mock_manage_handler, mock_get_handler, mock_myformatter):
        mock_manage_handler.return_value.__enter__ = mock.Mock()
        mock_manage_handler.return_value.__exit__ = mock.Mock()

        task.logger = mock.Mock()
        task.logger.info = mock.Mock()
        task.logger.log = mock.Mock()

        get_handler_inst1 = mock.Mock()
        get_handler_inst1.setFormatter = mock.Mock()
        get_handler_inst2 = mock.Mock()
        get_handler_inst2.setFormatter = mock.Mock()
        mock_get_handler.side_effect = [get_handler_inst1, get_handler_inst2]

        myformatter_inst1 = mock.Mock()
        myformatter_inst2 = mock.Mock()
        mock_myformatter.side_effect = [myformatter_inst1, myformatter_inst2]

        self.task_check_inst.messages = [("info", "info message"), ("warn", "warn message")]
        utils.CGREEN = ""
        utils.CEND = ""
        utils.CRED = ""

        self.task_check_inst.to_stdout()

        mock_manage_handler.assert_called_once_with("file", keep=False)
        mock_get_handler.assert_has_calls([
            mock.call(task.logger, "stream"),
            mock.call(task.logger, "stream")
            ])
        get_handler_inst1.setFormatter.assert_called_once_with(myformatter_inst1)
        get_handler_inst2.setFormatter.assert_called_once_with(myformatter_inst2)
        mock_myformatter.assert_has_calls([
            mock.call(flush=False),
            mock.call()
            ])
        task.logger.info.assert_called_once_with('task check job1 [Pass]', extra={'timestamp': '[2019/07/10 01:15:15]'})
        task.logger.log.assert_has_calls([
            mock.call(20, 'info message', extra={'timestamp': '  '}).
            mock.call(30, 'warn message', extra={'timestamp': '  '})
            ])

    @mock.patch('crmsh.crash_test.utils.json_dumps')
    def test_to_json(self, mock_dumps):
        self.task_check_inst.build_base_result = mock.Mock()
        self.task_check_inst.result = {}
        self.task_check_inst.to_json()
        self.task_check_inst.build_base_result.assert_called_once_with()
        mock_dumps.assert_called_once_with()

    def test_print_result(self):
        self.task_check_inst.to_stdout = mock.Mock()
        self.task_check_inst.to_json = mock.Mock()
        self.task_check_inst.print_result()
        self.task_check_inst.to_stdout.assert_called_once_with()
        self.task_check_inst.to_json.assert_called_once_with()

    def test_print_result_quiet(self):
        self.task_check_inst.quiet = True
        self.task_check_inst.to_stdout = mock.Mock()
        self.task_check_inst.print_result()
        self.task_check_inst.to_stdout.assert_not_called()

    def test_run(self):
        self.task_check_inst.print_result = mock.Mock()
        with self.task_check_inst.run():
            pass
        self.task_check_inst.print_result.assert_called_once_with()


class TestTaskSplitBrain(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    @mock.patch('crmsh.crash_test.utils.msg_info')
    def setUp(self, mock_msg_info):
        """
        Test setUp.
        """
        self.task_sp_inst = task.TaskSplitBrain()
        self.task_sp_inst.fence_action = "reboot"
        self.task_sp_inst.fence_timeout = 60

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    def test_header(self):
        expected_res = """==============================================
Testcase:          Simulate split brain by blocking traffic between cluster nodes
Expected Result:   One of nodes get fenced
Fence action:      reboot
Fence timeout:     60
"""
        res = self.task_sp_inst.header()
        self.assertEqual(res, expected_res)

    @mock.patch('crmsh.crash_test.utils.json_dumps')
    @mock.patch('crmsh.crash_test.task.Task.build_base_result')
    def test_to_json(self, mock_result, mock_json):
        self.task_sp_inst.result = {}
        self.task_sp_inst.to_json()
        mock_result.assert_called_once_with()
        mock_json.assert_called_once_with()

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.Task.task_pre_check')
    def test_pre_check_no_cmd(self, mock_pre_check, mock_run):
        mock_run.return_value = (1, None, "error")
        with self.assertRaises(task.TaskError) as err:
            self.task_sp_inst.pre_check()
        self.assertEqual("error", str(err.exception))
        mock_run.assert_called_once_with("which iptables")
        mock_pre_check.assert_called_once_with()

    @mock.patch('crmsh.crash_test.utils.online_nodes')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.Task.task_pre_check')
    def test_pre_check_error(self, mock_pre_check, mock_run, mock_online_nodes):
        mock_run.return_value = (0, None, None)
        mock_online_nodes.return_value = ["node1"]
        with self.assertRaises(task.TaskError) as err:
            self.task_sp_inst.pre_check()
        self.assertEqual("At least two nodes online!", str(err.exception))
        mock_run.assert_called_once_with("which iptables")
        mock_online_nodes.assert_called_once_with()

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.crmshutils.get_iplist_from_name')
    @mock.patch('crmsh.crash_test.task.Task.info')
    @mock.patch('crmsh.crash_test.utils.peer_node_list')
    def test_do_block_iptables(self, mock_peer_list, mock_info, mock_get_iplist, mock_run):
        mock_peer_list.return_value = ["node1", "node2"]
        mock_get_iplist.side_effect = [["10.10.10.1", "20.20.20.1"], ["10.10.10.2", "20.20.20.2"]]
        self.task_sp_inst.do_block_iptables()
        mock_peer_list.assert_called_once_with()
        mock_info.assert_has_calls([
            mock.call("Trying to temporarily block node1 communication ip"),
            mock.call("Trying to temporarily block node2 communication ip")
            ])
        mock_get_iplist.assert_has_calls([
            mock.call("node1"),
            mock.call("node2")
            ])
        mock_run.assert_has_calls([
            mock.call(config.BLOCK_IP.format(action='I', peer_ip="10.10.10.1")),
            mock.call(config.BLOCK_IP.format(action='I', peer_ip="20.20.20.1")),
            mock.call(config.BLOCK_IP.format(action='I', peer_ip="10.10.10.2")),
            mock.call(config.BLOCK_IP.format(action='I', peer_ip="20.20.20.2"))
            ])

    @mock.patch('crmsh.crash_test.task.TaskSplitBrain.un_block_iptables')
    def test_un_block(self, mock_unblock_iptables):
        self.task_sp_inst.un_block()
        mock_unblock_iptables.assert_called_once_with()

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.crmshutils.get_iplist_from_name')
    @mock.patch('crmsh.crash_test.task.Task.info')
    def test_un_block_iptables(self, mock_info, mock_get_iplist, mock_run):
        mock_get_iplist.side_effect = [["10.10.10.1", "20.20.20.1"], ["10.10.10.2", "20.20.20.2"]]
        self.task_sp_inst.peer_nodelist = ["node1", "node2"]
        self.task_sp_inst.un_block_iptables()
        mock_info.assert_has_calls([
            mock.call("Trying to recover node1 communication ip"),
            mock.call("Trying to recover node2 communication ip")
            ])
        mock_get_iplist.assert_has_calls([
            mock.call("node1"),
            mock.call("node2")
            ])
        mock_run.assert_has_calls([
            mock.call(config.BLOCK_IP.format(action='D', peer_ip="10.10.10.1")),
            mock.call(config.BLOCK_IP.format(action='D', peer_ip="20.20.20.1")),
            mock.call(config.BLOCK_IP.format(action='D', peer_ip="10.10.10.2")),
            mock.call(config.BLOCK_IP.format(action='D', peer_ip="20.20.20.2"))
            ])

    @mock.patch('crmsh.crash_test.task.Task.fence_action_monitor')
    @mock.patch('threading.Thread')
    def test_run(self, mock_thread, mock_monitor):
        mock_thread_inst = mock.Mock()
        mock_thread.return_value = mock_thread_inst
        self.task_sp_inst.run()
        mock_thread.assert_called_once_with(target=mock_monitor)
        mock_thread_inst.start.assert_called_once_with()

    def test_wait(self):
        self.task_sp_inst.fence_finish_event = mock.Mock()
        self.task_sp_inst.fence_finish_event.wait.return_value = False
        self.task_sp_inst.thread_stop_event = mock.Mock()
        self.task_sp_inst.wait()
        self.task_sp_inst.fence_finish_event.wait.assert_called_once_with(60)
        self.task_sp_inst.thread_stop_event.set.assert_called_once_with()


class TestFence(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    @mock.patch('crmsh.crash_test.utils.msg_info')
    def setUp(self, mock_msg_info):
        """
        Test setUp.
        """
        ctx = mock.Mock(fence_node="node1", yes=False)
        self.task_fence_inst = task.TaskFence(ctx)
        self.task_fence_inst.fence_action = "reboot"
        self.task_fence_inst.fence_timeout = 60

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    def test_header(self):
        expected_res = """==============================================
Testcase:          Fence node node1
Fence action:      reboot
Fence timeout:     60
"""
        res = self.task_fence_inst.header()
        self.assertEqual(res, expected_res)

    @mock.patch('crmsh.crash_test.utils.json_dumps')
    @mock.patch('crmsh.crash_test.task.Task.build_base_result')
    def test_to_json(self, mock_result, mock_json):
        self.task_fence_inst.result = {}
        self.task_fence_inst.to_json()
        mock_result.assert_called_once_with()
        mock_json.assert_called_once_with()

    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.Task.task_pre_check')
    def test_pre_check_no_cmd(self, mock_pre_check, mock_run):
        mock_run.return_value = (1, None, "error")
        with self.assertRaises(task.TaskError) as err:
            self.task_fence_inst.pre_check()
        self.assertEqual("error", str(err.exception))
        mock_run.assert_called_once_with("which crm_node")
        mock_pre_check.assert_called_once_with()

    @mock.patch('crmsh.crash_test.utils.check_node_status')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.Task.task_pre_check')
    def test_pre_check_error(self, mock_pre_check, mock_run, mock_node_status):
        mock_run.side_effect = [(0, None, None), (0, None, None), (0, None, None)]
        mock_node_status.return_value = False
        with self.assertRaises(task.TaskError) as err:
            self.task_fence_inst.pre_check()
        self.assertEqual("Node \"node1\" not in cluster!", str(err.exception))
        mock_run.assert_has_calls([
            mock.call("which crm_node"),
            mock.call("which stonith_admin"),
            mock.call("which crm_attribute")
            ])
        mock_node_status.assert_called_once_with("node1", "member")

    @mock.patch('crmsh.crash_test.task.Task.fence_action_monitor')
    @mock.patch('threading.Thread')
    @mock.patch('crmsh.sh.ShellUtils.get_stdout_stderr')
    @mock.patch('crmsh.crash_test.task.Task.info')
    def test_run(self, mock_info, mock_run, mock_thread, mock_monitor):
        mock_thread_inst = mock.Mock()
        mock_thread.return_value = mock_thread_inst
        self.task_fence_inst.run()
        mock_info.assert_called_once_with("Trying to fence node \"node1\"")
        mock_run.assert_called_once_with("crm_attribute -t status -N 'node1' -n terminate -v true")
        mock_thread.assert_called_once_with(target=mock_monitor)
        mock_thread_inst.start.assert_called_once_with()

    @mock.patch('crmsh.crash_test.task.Task.info')
    @mock.patch('crmsh.crash_test.utils.this_node')
    def test_wait_this_node(self, mock_this_node, mock_info):
        mock_this_node.return_value = "node1"
        self.task_fence_inst.fence_finish_event = mock.Mock()
        self.task_fence_inst.thread_stop_event = mock.Mock()
        self.task_fence_inst.fence_finish_event.wait.return_value = True

        self.task_fence_inst.wait()

        mock_this_node.assert_called_once_with()
        mock_info.assert_called_once_with("Waiting 60s for self reboot...")
        self.task_fence_inst.fence_finish_event.wait.assert_called_once_with(60)

    @mock.patch('crmsh.crash_test.task.Task.info')
    @mock.patch('crmsh.crash_test.utils.this_node')
    def test_wait(self, mock_this_node, mock_info):
        mock_this_node.return_value = "node2"
        self.task_fence_inst.fence_finish_event = mock.Mock()
        self.task_fence_inst.thread_stop_event = mock.Mock()
        self.task_fence_inst.fence_finish_event.wait.return_value = None

        with self.assertRaises(task.TaskError) as err:
            self.task_fence_inst.wait()
        self.assertEqual("Target fence node \"node1\" still alive", str(err.exception))

        mock_this_node.assert_called_once_with()
        mock_info.assert_called_once_with("Waiting 60s for node \"node1\" reboot...")
        self.task_fence_inst.fence_finish_event.wait.assert_called_once_with(60)


class TestTask(TestCase):

    @classmethod
    def setUpClass(cls):
        """
        Global setUp.
        """

    @mock.patch('crmsh.crash_test.utils.msg_info')
    def setUp(self, mock_info):
        """
        Test setUp.
        """
        fake_now = datetime(2019, 7, 10, 1, 15, 15)
        main.ctx = mock.Mock(task_list={"process_name": "xin", "age": 38})
        self.task_inst = task.Task("task description", flush=True)
        self.task_inst.timestamp = fake_now

    def tearDown(self):
        """
        Test tearDown.
        """

    @classmethod
    def tearDownClass(cls):
        """
        Global tearDown.
        """

    def test_header(self):
        self.task_inst.header()

    def test_to_report(self):
        self.task_inst.to_report()

    def test_to_json(self):
        self.task_inst.to_json()

    @mock.patch('crmsh.service_manager.ServiceManager.service_is_active')
    def test_task_pre_check_exception(self, mock_active):
        mock_active.return_value = False
        with self.assertRaises(task.TaskError) as err:
            self.task_inst.task_pre_check()
        self.assertEqual("Cluster not running!", str(err.exception))
        mock_active.assert_called_once_with("pacemaker.service")

    @mock.patch('crmsh.service_manager.ServiceManager.service_is_active')
    def test_task_pre_check_exception_no_fence(self, mock_active):
        mock_active.return_value = True
        self.task_inst.get_fence_info = mock.Mock()
        self.task_inst.fence_enabled = False
        with self.assertRaises(task.TaskError) as err:
            self.task_inst.task_pre_check()
        self.assertEqual("Require stonith enabled", str(err.exception))
        mock_active.assert_called_once_with("pacemaker.service")
        self.task_inst.get_fence_info.assert_called_once_with()

    @mock.patch('crmsh.crash_test.utils.FenceInfo')
    def test_get_fence_info(self, mock_fence_info):
        mock_fence_info_inst = mock.Mock()
        mock_fence_info.return_value = mock_fence_info_inst
        self.task_inst.get_fence_info()

    @mock.patch('crmsh.crash_test.utils.msg_info')
    def test_info(self, mock_info):
        self.task_inst.msg_append = mock.Mock()
        self.task_inst.info("info message")
        self.task_inst.msg_append.assert_called_once_with("info", "info message")
        mock_info.assert_called_once_with("info message", to_stdout=True)

    @mock.patch('crmsh.crash_test.utils.msg_warn')
    def test_warn(self, mock_warn):
        self.task_inst.msg_append = mock.Mock()
        self.task_inst.warn("warn message")
        self.task_inst.msg_append.assert_called_once_with("warn", "warn message")
        mock_warn.assert_called_once_with("warn message", to_stdout=True)

    @mock.patch('crmsh.crash_test.utils.msg_error')
    def test_error(self, mock_error):
        self.task_inst.msg_append = mock.Mock()
        self.task_inst.error("error message")
        self.task_inst.msg_append.assert_called_once_with("error", "error message")
        mock_error.assert_called_once_with("error message", to_stdout=True)

    @mock.patch('crmsh.crash_test.utils.now')
    def test_msg_append(self, mock_now):
        self.task_inst.to_json = mock.Mock()
        self.task_inst.to_report = mock.Mock()
        self.task_inst.msg_append("error", "warn message")
        mock_now.assert_called_once_with()
        self.task_inst.to_json.assert_called_once_with()
        self.task_inst.to_report.assert_called_once_with()

    def test_build_base_result(self):
        self.task_inst.build_base_result()
        expected_result = {
            "Timestamp": '2019/07/10 01:15:15',
            "Description": self.task_inst.description,
            "Messages": []
        }
        self.assertDictEqual(expected_result, self.task_inst.result)

    @mock.patch('crmsh.crash_test.utils.warning_ask')
    def test_print_header(self, mock_ask):
        self.task_inst.header = mock.Mock()
        self.task_inst.info = mock.Mock()
        mock_ask.return_value = False

        with self.assertRaises(crmshutils.TerminateSubCommand):
            self.task_inst.print_header()

        self.task_inst.header.assert_called_once_with()
        mock_ask.assert_called_once_with(task.Task.REBOOT_WARNING)
        self.task_inst.info.assert_called_once_with("Testcase cancelled")

    @mock.patch('time.sleep')
    @mock.patch('crmsh.crash_test.task.Task.info')
    @mock.patch('crmsh.xmlutil.CrmMonXmlParser')
    def test_fence_action_monitor(self, mock_parser, mock_info, mock_sleep):
        self.task_inst.thread_stop_event = mock.Mock()
        self.task_inst.thread_stop_event.is_set.side_effect = [False, False]
        self.task_inst.fence_start_event = mock.Mock()
        self.task_inst.fence_start_event.is_set.side_effect = [False, True]
        self.task_inst.fence_finish_event = mock.Mock()
        mock_parser_inst = mock.Mock()
        mock_parser.return_value = mock_parser_inst
        mock_parser_inst.get_last_fence_event_info.side_effect = [
            {"target": "node2", "origin": "node1", "status": "pending", "completed": ""},
            {"target": "node2", "origin": "node1", "status": "success", "completed": "2025-05-30 10:41:58.376958 +08:00"},
        ]

        self.task_inst.fence_action_monitor()

        self.task_inst.thread_stop_event.is_set.assert_has_calls([mock.call(), mock.call()])
        mock_info.assert_has_calls([
            mock.call("Node \"node2\" will be fenced by \"node1\"!"),
            mock.call("Node \"node2\" was fenced by \"node1\" at 2025-05-30 10:41:58.376958 +08:00")
            ])
        self.task_inst.fence_start_event.set.assert_called_once_with()
        self.task_inst.fence_finish_event.set.assert_called_once_with()


class TestFixSBD(TestCase):
    """
    Class to test TaskFixSBD of task.py
    All tested in test_crash_test.py except verify()
    """

    @mock.patch('builtins.open')
    @mock.patch('os.path.isfile')
    @mock.patch('tempfile.mkstemp')
    @mock.patch('crmsh.crash_test.utils.msg_info')
    def setUp(self, mock_msg_info, mock_mkstemp, mock_isfile, mock_open):
        """
        Test setUp.
        """
        dev = "/dev/disk/by-id/scsi-SATA_ST2000LM007-1R81_WDZ5J42A"
        bak = "/tmp/tmpmby3ty9g"
        edit = "/tmp/tmpnic4t30s"
        mock_isfile.return_value = True
        mock_open.return_value = mock.mock_open(read_data="SBD_DEVICE={}".
                                                format(dev)).return_value
        mock_mkstemp.side_effect = [(1, bak), (2, edit)]

        self.task_fixsbd = task.TaskFixSBD(dev, force=False)
        mock_msg_info.assert_called_once_with('Replace SBD_DEVICE with candidate {}'.
                                              format(dev), to_stdout=False)

    def tearDown(self):
        """
        Test tearDown.
        """
        pass

    @mock.patch('os.fsync')
    @mock.patch('builtins.open')
    @mock.patch('os.path.isfile')
    @mock.patch('crmsh.crash_test.utils.msg_info')
    def test_verify_succeed(self, mock_msg_info, mock_isfile, mock_open, mock_fsync):
        """
        Test verify successful.
        """
        dev = "/dev/disk/by-id/scsi-SATA_ST2000LM007-1R81_WDZ5J42A"
        mock_isfile.return_value = True
        mock_open.return_value = mock.mock_open(read_data="SBD_DEVICE={}".
                                                format(dev)).return_value
        self.task_fixsbd.prev_task_list = []

        self.task_fixsbd.verify()
        mock_isfile.assert_called_once_with(config.SBD_CONF)
        mock_msg_info.assert_called_once_with('SBD DEVICE change succeed',
                                              to_stdout=True)
        mock_fsync.assert_called()

    @mock.patch('builtins.open')
    @mock.patch('os.path.isfile')
    def test_verify_fail(self, mock_isfile, mock_open):
        """
        Test verify failed.
        """
        dev = "/dev/disk/by-id/scsi-SATA_ST2000LM007-1R81_WDZ5J42A"
        dev_cur = "/dev/disk/by-id/scsi-SATA_ST2000LM007-no_change"
        mock_isfile.return_value = True
        mock_open.return_value = mock.mock_open(read_data="SBD_DEVICE={}".
                                                format(dev_cur)).return_value
        self.task_fixsbd.prev_task_list = []

        with self.assertRaises(task.TaskError) as err:
            self.task_fixsbd.verify()
        mock_isfile.assert_called_once_with(config.SBD_CONF)
        self.assertEqual("Fail to replace SBD device {} in {}!".
                         format(dev, config.SBD_CONF), str(err.exception))



import os
import sys
import getpass
import select
import subprocess as proc
try:
    import json
except ImportError:
    import simplejson as json

_input = None

# read stdin, if there's anything to read
_stdin_data = {}
while sys.stdin in select.select([sys.stdin], [], [], 0)[0]:
    line = sys.stdin.readline()
    if line:
        d = line.split(':', 1)
        if len(d) == 2:
            _stdin_data[d[0].strip()] = d[1].strip()
    else:
        break


def decode_utf8(s):
    """
    Convert the given byte sequence to a
    utf8 string.
    """
    if s is None or isinstance(s, str):
        return s
    return s.decode('utf-8', 'ignore')


def host():
    return os.uname()[1]


def get_input():
    global _input
    if _input is None:
        _input = json.load(open('./script.input'))
    return _input


def parameters():
    return get_input()[0]


def param(name):
    return parameters().get(name)


def output(step_idx):
    if step_idx < len(get_input()):
        return get_input()[step_idx]
    return {}


def exit_fail(msg):
    print(msg, file=sys.stderr)
    sys.exit(1)


def exit_ok(data):
    print(json.dumps(data))
    sys.exit(0)


def is_true(s):
    if s in (True, False):
        return s
    return isinstance(s, str) and s.lower() in ('yes', 'true', '1', 'on')


_debug_enabled = None


def debug_enabled():
    global _debug_enabled
    if _debug_enabled is None:
        _debug_enabled = is_true(param('debug'))
    return _debug_enabled


def info(msg):
    "writes msg to log"
    with open('./crm_script.debug', 'a') as dbglog:
        dbglog.write('%s' % (msg))


def debug(msg):
    "writes msg to log and syslog if debug is enabled"
    if debug_enabled():
        try:
            with open('./crm_script.debug', 'a') as dbglog:
                dbglog.write('%s\n' % (msg))
            import syslog
            syslog.openlog("crmsh", 0, syslog.LOG_USER)
            syslog.syslog(syslog.LOG_NOTICE, str(msg).encode('utf8'))
        except:
            pass


def call(cmd, shell=False):
    debug("crm_script(call): %s" % (cmd))
    p = proc.Popen(cmd, shell=shell, stdin=None, stdout=proc.PIPE, stderr=proc.PIPE)
    out, err = p.communicate()
    return p.returncode, decode_utf8(out).strip(), decode_utf8(err).strip()


def use_sudo():
    return getpass.getuser() != 'root' and is_true(param('sudo')) and _stdin_data.get('sudo')


def sudo_call(cmd, shell=False):
    if not use_sudo():
        return call(cmd, shell=shell)
    debug("crm_script(sudo_call): %s" % (cmd))
    os.unsetenv('SSH_ASKPASS')
    call(['sudo', '-k'], shell=False)
    sudo_prompt = 'crm_script_sudo_prompt'
    if isinstance(cmd, str):
        cmd = "sudo -H -S -p '%s' %s" % (sudo_prompt, cmd)
    else:
        cmd = ['sudo', '-H', '-S', '-p', sudo_prompt] + cmd
    p = proc.Popen(cmd, shell=shell, stdin=proc.PIPE, stdout=proc.PIPE, stderr=proc.PIPE)
    sudo_pass = "%s\n" % (_stdin_data.get('sudo', 'linux'))
    debug("CMD(SUDO): %s" % (str(cmd)))
    out, err = p.communicate(input=sudo_pass)
    return p.returncode, out.strip(), err.strip()


def service(name, action):
    if action.startswith('is-'):
        return call(['/usr/bin/systemctl', action, name + '.service'])
    return sudo_call(['/usr/bin/systemctl', action, name + '.service'])


def package(name, state):
    rc, out, err = sudo_call(['./crm_pkg.py', '-n', name, '-s', state])
    if rc != 0:
        raise IOError("%s / %s" % (out, err))
    outp = json.loads(decode_utf8(out))
    if isinstance(outp, dict) and 'rc' in outp:
        rc = int(outp['rc'])
        if rc != 0:
            raise IOError("(rc=%s) %s%s" % (rc, outp.get('stdout', ''), outp.get('stderr', '')))
    return outp


def check_package(name, state):
    rc, out, err = call(['./crm_pkg.py', '--dry-run', '-n', name, '-s', state])
    if rc != 0:
        raise IOError(err)
    outp = json.loads(out)
    if isinstance(outp, dict) and 'rc' in outp:
        rc = int(outp['rc'])
        if rc != 0:
            raise IOError("(rc=%s) %s%s" % (rc, outp.get('stdout', ''), outp.get('stderr', '')))
    return outp


def rpmcheck(names):
    rc, out, err = call(['./crm_rpmcheck.py'] + names)
    if rc != 0:
        raise IOError(err)
    return json.loads(out)


def save_template(template, dest, **kwargs):
    '''
    1. Reads a template from <template>,
    2. Replaces all template variables with those in <kwargs> and
    3. writes the resulting file to <dest>
    '''
    import re
    tmpl = open(template).read()
    keys = re.findall(r'%\((\w+)\)s', tmpl, re.MULTILINE)
    missing_keys = set(keys) - set(kwargs.keys())
    if missing_keys:
        raise ValueError("Missing template arguments: %s" % ', '.join(missing_keys))
    tmpl = tmpl % kwargs
    try:
        with open(dest, 'w') as f:
            f.write(tmpl)
    except Exception as e:
        raise IOError("Failed to write %s from template %s: %s" % (dest, template, e))
    debug("crm_script(save_template): wrote %s" % (dest))




CRM_H_OUTPUT = '''usage: crm [-h|--help] [OPTIONS] [SUBCOMMAND ARGS...]
or crm help SUBCOMMAND

For a list of available subcommands, use crm help.

Use crm without arguments for an interactive session.
Call a subcommand directly for a "single-shot" use.
Call crm with a level name as argument to start an interactive
session from that level.

See the crm(8) man page or call crm help for more details.

positional arguments:
  SUBCOMMAND

options:
  -h, --help            show this help message and exit
  --version             show program's version number and exit
  -f, --file FILE       Load commands from the given file. If a dash (-) is
                        used in place of a file name, crm will read commands
                        from the shell standard input (stdin).
  -c, --cib CIB         Start the session using the given shadow CIB file.
                        Equivalent to `cib use <CIB>`.
  -D, --display OUTPUT_TYPE
                        Choose one of the output options: plain, color-always,
                        color, or uppercase. The default is color if the
                        terminal emulation supports colors, else plain.
  -F, --force           Make crm proceed with applying changes where it would
                        normally ask the user to confirm before proceeding.
                        This option is mainly useful in scripts, and should be
                        used with care.
  -n, --no              Automatically answer no when prompted
  -w, --wait            Make crm wait for the cluster transition to finish
                        (for the changes to take effect) after each processed
                        line.
  -H, --history DIR|FILE|SESSION
                        A directory or file containing a cluster report to
                        load into history, or the name of a previously saved
                        history session.
  -d, --debug           Print verbose debugging information.
  -R, --regression-tests
                        Enables extra verbose trace logging used by the
                        regression tests. Logs all external calls made by
                        crmsh.
  --scriptdir DIR       Extra directory where crm looks for cluster scripts,
                        or a list of directories separated by semi-colons
                        (e.g. /dir1;/dir2;etc.).
  -X PROFILE            Collect profiling data and save in PROFILE.
  -o, --opt OPTION=VALUE
                        Set crmsh option temporarily. If the options are saved
                        using+options save+ then the value passed here will
                        also be saved.Multiple options can be set by using
                        +-o+ multiple times.'''


CRM_CLUSTER_INIT_H_OUTPUT = '''Initializes a new HA cluster

usage: init [options] [STAGE]

Initialize a cluster from scratch. This command configures
a complete cluster, and can also add additional cluster
nodes to the initial one-node cluster using the --nodes
option.

options:
  -h, --help            Show this help message
  -q, --quiet           Be quiet (don't describe what's happening, just do it)
  -y, --yes             Answer "yes" to all prompts (use with caution, this is
                        destructive, especially those storage related
                        configurations and stages.)
  -n, --name NAME       Set the name of the configured cluster.
  -N, --node [USER@]HOST
                        The member node of the cluster. Note: the current node
                        is always get initialized during bootstrap in the
                        beginning.
  -S, --enable-sbd      Enable SBD even if no SBD device is configured
                        (diskless mode)
  -w, --watchdog WATCHDOG
                        Use the given watchdog device or driver name
  -x, --skip-csync2-sync
                        Skip csync2 initialization (an experimental option)
  --use-ssh-agent, --no-use-ssh-agent
                        Try to use an existing key from ssh-agent (default)

Network configuration:
  Options for configuring the network and messaging layer.

  -i, --interface IF    Bind to IP address on interface IF. Allowed value is
                        nic name or IP address. If a nic name is provided, the
                        first IP of that nic will be used. Use multiple -i for
                        more links. Note: Only one link is allowed for the non
                        knet transport type
  -t, --transport TRANSPORT
                        The transport mechanism. Allowed value is
                        knet(kronosnet)/udpu(unicast)/udp(multicast). Default
                        is knet
  -A, --admin-ip IP     Configure IP address as an administration virtual IP
  -I, --ipv6            Configure corosync use IPv6

QDevice configuration:
  QDevice participates in quorum decisions. With the assistance of 
  a third-party arbitrator Qnetd, it provides votes so that a cluster 
  is able to sustain more node failures than standard quorum rules 
  allow. It is recommended for clusters with an even number of nodes 
  and highly recommended for 2 node clusters.
  
  Options for configuring QDevice and QNetd.

  --qnetd-hostname [USER@]HOST
                        User and host of the QNetd server. The host can be
                        specified in either hostname or IP address.
  --qdevice-port PORT   TCP PORT of QNetd server (default:5403)
  --qdevice-algo ALGORITHM
                        QNetd decision ALGORITHM (ffsplit/lms,
                        default:ffsplit)
  --qdevice-tie-breaker TIE_BREAKER
                        QNetd TIE_BREAKER (lowest/highest/valid_node_id,
                        default:lowest)
  --qdevice-tls TLS     Whether using TLS on QDevice (on/off/required,
                        default:on)
  --qdevice-heuristics COMMAND
                        COMMAND to run with absolute path. For multiple
                        commands, use ";" to separate (details about
                        heuristics can see man 8 corosync-qdevice)
  --qdevice-heuristics-mode MODE
                        MODE of operation of heuristics (on/sync/off,
                        default:sync)

Storage configuration:
  Options for configuring shared storage.

  -s, --sbd-device DEVICE
                        Block device to use for SBD fencing, use ";" as
                        separator or -s multiple times for multi path (up to 3
                        devices)
  -o, --ocfs2-device DEVICE
                        Block device to use for OCFS2; When using Cluster LVM2
                        to manage the shared storage, user can specify one or
                        multiple raw disks, use ";" as separator or -o
                        multiple times for multi path (must specify -C option)
                        NOTE: this is a Technical Preview
  -g, --gfs2-device DEVICE
                        Block device to use for GFS2; When using Cluster LVM2
                        to manage the shared storage, user can specify one or
                        multiple raw disks, use ";" as separator or -g
                        multiple times for multi path (must specify -C option)
                        NOTE: this is a Technical Preview
  -C, --cluster-lvm2    Use Cluster LVM2 (only valid together with -o or -g
                        option) NOTE: this is a Technical Preview
  -m, --mount-point MOUNT
                        Mount point for OCFS2 or GFS2 device (default is
                        /srv/clusterfs, only valid together with -o or -g
                        option) NOTE: this is a Technical Preview

Stage can be one of:
    ssh         Create SSH keys for passwordless SSH between cluster nodes
    firewalld   Add high-availability service to firewalld
    csync2      Configure csync2
    corosync    Configure corosync
    sbd         Configure SBD (requires -s <dev>)
    cluster     Bring the cluster online
    ocfs2       Configure OCFS2 (requires -o <dev>) NOTE: this is a Technical Preview
    gfs2        Configure GFS2 (requires -g <dev>) NOTE: this is a Technical Preview
    admin       Create administration virtual IP (optional)
    qdevice     Configure qdevice and qnetd

Note:
  - If stage is not specified, the script will run through each stage
    in sequence, with prompts for required information.

Examples:
  # Setup the cluster on the current node
  crm cluster init -y

  # Setup the cluster with multiple nodes
  (NOTE: the current node will be part of the cluster even not listed in the -N option as below)
  crm cluster init -N node1 -N node2 -N node3 -y

  # Setup the cluster on the current node, with two network interfaces
  crm cluster init -i eth1 -i eth2 -y

  # Setup the cluster on the current node, with disk-based SBD
  crm cluster init -s <share disk> -y

  # Setup the cluster on the current node, with diskless SBD
  crm cluster init -S  -y

  # Setup the cluster on the current node, with QDevice
  crm cluster init --qnetd-hostname <qnetd addr> -y

  # Setup the cluster on the current node, with SBD+OCFS2
  crm cluster init -s <share disk1> -o <share disk2> -y

  # Setup the cluster on the current node, with SBD+GFS2
  crm cluster init -s <share disk1> -g <share disk2> -y

  # Setup the cluster on the current node, with SBD+OCFS2+Cluster LVM
  crm cluster init -s <share disk1> -o <share disk2> -o <share disk3> -C -y

  # Setup the cluster on the current node, with SBD+GFS2+Cluster LVM
  crm cluster init -s <share disk1> -g <share disk2> -g <share disk3> -C -y

  # Add SBD on a running cluster
  crm cluster init sbd -s <share disk> -y

  # Replace SBD device on a running cluster which already configured SBD
  crm -F cluster init sbd -s <share disk> -y

  # Add diskless SBD on a running cluster
  crm cluster init sbd -S -y

  # Add QDevice on a running cluster
  crm cluster init qdevice --qnetd-hostname <qnetd addr> -y

  # Add OCFS2+Cluster LVM on a running cluster
  crm cluster init ocfs2 -o <share disk1> -o <share disk2> -C -y

  # Add GFS2+Cluster LVM on a running cluster
  crm cluster init gfs2 -g <share disk1> -g <share disk2> -C -y'''


CRM_CLUSTER_JOIN_H_OUTPUT = '''Join existing cluster

usage: join [options] [STAGE]

Join the current node to an existing cluster. The
current node cannot be a member of a cluster already.
Pass any node in the existing cluster as the argument
to the -c option.

options:
  -h, --help            Show this help message
  -q, --quiet           Be quiet (don't describe what's happening, just do it)
  -y, --yes             Answer "yes" to all prompts (use with caution)
  --use-ssh-agent, --no-use-ssh-agent
                        Try to use an existing key from ssh-agent (default)

Network configuration:
  Options for configuring the network and messaging layer.

  -c, --cluster-node [USER@]HOST
                        User and host to login to an existing cluster node.
                        The host can be specified with either a hostname or an
                        IP.
  -i, --interface IF    Bind to IP address on interface IF. Allowed value is
                        nic name or IP address. If a nic name is provided, the
                        first IP of that nic will be used. Use multiple -i for
                        more links. Note: Only one link is allowed for the non
                        knet transport type

Stage can be one of:
    ssh         Obtain SSH keys from existing cluster node (requires -c <host>)
    firewalld   Add high-availability service to firewalld
    csync2      Configure csync2 (requires -c <host>)
    ssh_merge   Merge root's SSH known_hosts across all nodes (csync2 must
                already be configured).
    cluster     Start the cluster on this node

If stage is not specified, each stage will be invoked in sequence.

Examples:
  # Join with a cluster node
  crm cluster join -c <node> -y

  # Join with a cluster node, with the same network interface used by that node
  crm cluster join -c <node> -i eth1 -i eth2 -y'''


CRM_CLUSTER_REMOVE_H_OUTPUT = '''Remove node(s) from the cluster

usage: remove [options] [<node> ...]

Remove one or more nodes from the cluster.

This command can remove the last node in the cluster,
thus effectively removing the whole cluster. To remove
the last node, pass --force argument to crm or set
the config.core.force option.

options:
  -h, --help            Show this help message
  -q, --quiet           Be quiet (don't describe what's happening, just do it)
  -y, --yes             Answer "yes" to all prompts (use with caution)
  -c, --cluster-node HOST
                        IP address or hostname of cluster node which will be
                        deleted
  -F, --force           Remove current node
  --qdevice             Remove QDevice configuration and service from cluster'''


CRM_CLUSTER_GEO_INIT_H_OUTPUT = '''Configure cluster as geo cluster

usage: geo-init [options]

Create a new geo cluster with the current cluster as the
first member. Pass the complete geo cluster topology as
arguments to this command, and then use geo-join and
geo-init-arbitrator to add the remaining members to
the geo cluster.

options:
  -h, --help            Show this help message
  -q, --quiet           Be quiet (don't describe what's happening, just do it)
  -y, --yes             Answer "yes" to all prompts (use with caution)
  -a, --arbitrator [USER@]HOST
                        Geo cluster arbitrator
  -s, --clusters DESC   Geo cluster description (see details below)
  -t, --tickets LIST    Tickets to create (space-separated)

Cluster Description

  This is a map of cluster names to IP addresses.
  Each IP address will be configured as a virtual IP
  representing that cluster in the geo cluster
  configuration.

  Example with two clusters named paris and amsterdam:

  --clusters "paris=192.168.10.10 amsterdam=192.168.10.11"

  Name clusters using the --name parameter to
  crm bootstrap init.'''


CRM_CLUSTER_GEO_JOIN_H_OUTPUT = '''Join cluster to existing geo cluster

usage: geo-join [options]

This command should be run from one of the nodes in a cluster
which is currently not a member of a geo cluster. The geo
cluster configuration will be fetched from the provided node,
and the cluster will be added to the geo cluster.

Note that each cluster in a geo cluster needs to have a unique
name set. The cluster name can be set using the --name argument
to init, or by configuring corosync with the cluster name in
an existing cluster.

options:
  -h, --help            Show this help message
  -q, --quiet           Be quiet (don't describe what's happening, just do it)
  -y, --yes             Answer "yes" to all prompts (use with caution)
  -c, --cluster-node [USER@]HOST
                        An already-configured geo cluster or arbitrator
  -s, --clusters DESC   Geo cluster description (see geo-init for details)
  --use-ssh-agent, --no-use-ssh-agent
                        Try to use an existing key from ssh-agent (default)'''


CRM_CLUSTER_GEO_INIT_ARBIT_H_OUTPUT = '''Initialize node as geo cluster arbitrator

usage: geo-init-arbitrator [options]

Configure the current node as a geo arbitrator. The command
requires an existing geo cluster or geo arbitrator from which
to get the geo cluster configuration.

options:
  -h, --help            Show this help message
  -q, --quiet           Be quiet (don't describe what's happening, just do it)
  -y, --yes             Answer "yes" to all prompts (use with caution)
  -c, --cluster-node [USER@]HOST
                        An already-configured geo cluster
  --use-ssh-agent, --no-use-ssh-agent
                        Try to use an existing key from ssh-agent (default)'''

CRM_CONF_CONTENT_POSIONED = '''[core]
hosts = alan@hanode1, claude@hanode2, john@hanode3'''



from __future__ import unicode_literals
# Copyright (C) 2014 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.


from crmsh import cibconfig

factory = cibconfig.cib_factory


def assert_in(needle, haystack):
    if needle not in haystack:
        message = "%s not in %s" % (needle, haystack)
        raise AssertionError(message)


def setup_function():
    "set up test fixtures"
    from crmsh import idmgmt
    idmgmt.clear()


def teardown_function():
    pass


def test_nodes_nocli():
    for n in factory.node_id_list():
        obj = factory.find_object(n)
        if obj is not None:
            assert obj.node is not None
            assert True == obj.cli_use_validate()
            assert False == obj.nocli


def test_show():
    setobj = cibconfig.mkset_obj()
    s = setobj.repr_nopretty()
    sp = s.splitlines()
    assert_in("node ha-one", sp[0:3])



# Copyright (C) 2017 Xin Liang <XLiang@suse.com>
# See COPYING for license information.

import datetime
import glob
import os
import re
import shutil
import sys
import traceback
from dateutil import tz
from enum import Enum
from collections import Counter
from typing import Optional, List, Tuple

from crmsh import utils as crmutils
from crmsh import corosync, log, userdir, tmpfiles, config, sh
from crmsh.report import constants, collect, core
from crmsh.sh import ShellUtils


logger = log.setup_report_logger(__name__)


class LogType(Enum):
    GOOD = 0             # good log; include
    IRREGULAR = 1        # irregular log; include
    EMPTY = 2            # empty log; exclude
    BEFORE_TIMESPAN = 3  # log before timespan; exclude
    AFTER_TIMESPAN = 4   # log after timespan; exclude


def convert_logtype_to_str(log_type: LogType) -> str:
    log_type_str = {
        LogType.GOOD: "in timespan",
        LogType.IRREGULAR: "irregular",
        LogType.EMPTY: "empty",
        LogType.BEFORE_TIMESPAN: "before timespan",
        LogType.AFTER_TIMESPAN: "after timespan"
    }
    return log_type_str[log_type]


class ReportGenericError(Exception):
    pass


def arch_logs(context: core.Context, logf: str) -> Tuple[List[str], LogType]:
    """
    Go through archived logs and return those in timespan and the LogType
    """
    return_list = []
    log_type = None

    file_list = [logf] + glob.glob(logf+"*[0-9z]")
    do_have_archived_logs = len(file_list) > 1
    # like ls -t, newest first
    for f in sorted(file_list, key=os.path.getmtime, reverse=True):
        # for those archived logs, if the newest one is older than the from_time
        if do_have_archived_logs and context.from_time > os.path.getmtime(f):
            break # no need to check the rest
        tmp = is_our_log(context, f)
        logger.debug2("File %s is %s", f, convert_logtype_to_str(tmp))
        if tmp not in (LogType.GOOD, LogType.IRREGULAR):
            continue
        log_type = tmp
        return_list.append(f)

    if return_list:
        logger.debug2(
            "Found %s logs: %s",
            convert_logtype_to_str(log_type),
            ', '.join(return_list)
        )
    return return_list, log_type


def analyze(context: core.Context) -> None:
    """
    """
    result_list = []

    result_list.append(compare_and_consolidate_files(context))
    result_list += check_collected_files(context)
    result_list += extract_critical_log(context)

    analyze_f = os.path.join(context.work_dir, constants.ANALYSIS_F)
    crmutils.str2file('\n'.join(result_list), analyze_f)


def compare_and_consolidate_files(context: core.Context) -> str:
    out_string: str = ""
    workdir = context.work_dir
    compare_file_list = [
        constants.MEMBERSHIP_F,
        constants.CRM_MON_F,
        constants.COROSYNC_F,
        constants.SYSINFO_F,
        constants.CIB_F
    ]

    for f in compare_file_list:
        out_string += f"Diff {f}... "
        if not glob.glob(f"{workdir}/*/{f}"):
            out_string += f"no {f} found in {workdir}\n"
            continue
        rc, out = do_compare(context, f)
        out_string += f"\n{out}\n" if out else "OK\n"
        if rc == 0 and f != constants.CIB_F:
            consolidate(context, f)

    return out_string + '\n'


def do_compare(context: core.Context, file: str) -> Tuple[int, str]:
    """
    Compare file content between cluster nodes
    """
    rc, out_string = 0, ""
    prev_file_path = None

    for n in context.node_list:
        current_file_path = os.path.join(context.work_dir, n, file)

        if prev_file_path:
            rc, out = diff_check(prev_file_path, current_file_path)
            out_string += f"{out}\n" if out else ""
            rc += rc
        else:
            prev_file_path = current_file_path

    return rc, out_string


def check_collected_files(context: core.Context) -> List[str]:
    """
    Check collected files for warnings and issues
    """
    results = []
    file_description_dict = {
        constants.COREDUMP_F: "WARN: Coredump found at",
        constants.CRM_VERIFY_F: "WARN: crm_verify reported warnings at",
        constants.PERMISSIONS_F: "Checking problems with permissions/ownership at"
    }

    for node in context.node_list:
        for f, desc in file_description_dict.items():
            f_in_work_dir = os.path.join(context.work_dir, node, f)
            if os.path.isfile(f_in_work_dir) and not crmutils.file_is_empty(f_in_work_dir):
                results.append(f"{desc} {node}:")
                results.append(crmutils.read_from_file(f_in_work_dir))

    return results


def extract_critical_log(context: core.Context) -> List[str]:
    """
    Extract warnings and errors from collected log files
    """
    result_list = []
    grep_e_option_str = f"-e {' -e '.join(constants.LOG_PATTERNS)}"
    shell = sh.ShellUtils()

    for f in glob.glob(f"{context.work_dir}/*/*.log"):
        grep_cmd = f"grep -F {grep_e_option_str} {f}"
        _, out, _ = shell.get_stdout_stderr(grep_cmd)
        if out:
            result_list.append(f"\nWARNINGS or ERRORS in {real_path(f)}:")
            result_list.append(out)
    return result_list


def cib_diff(file1: str, file2: str) -> Tuple[int, str]:
    """
    check if CIB files have same content in the cluster
    """
    node1_dir = os.path.dirname(file1)
    node2_dir = os.path.dirname(file2)

    if (os.path.isfile(os.path.join(node1_dir, "RUNNING")) and
        os.path.isfile(os.path.join(node2_dir, "RUNNING"))) or \
        (os.path.isfile(os.path.join(node1_dir, "STOPPED")) and
         os.path.isfile(os.path.join(node2_dir, "STOPPED"))):
        cmd = f"crm_diff -c -n {file1} -o {file2}"
        code, out_string, _ = ShellUtils().get_stdout_stderr(cmd)
    else:
        code, out_string = 1, "Can't compare cibs from running and stopped systems\n"
    return code, out_string


def consolidate(context: core.Context, target_file: str) -> None:
    """
    Remove duplicates if files are same, make links instead
    """
    workdir = context.work_dir
    for node in context.node_list:
        target_file_in_path = os.path.join(workdir, node, target_file)
        if os.path.isfile(os.path.join(workdir, target_file)):
            os.remove(target_file_in_path)
        else:
            shutil.move(target_file_in_path, workdir)
        os.symlink(f"../{target_file}", target_file_in_path)


def diff_check(file1: str, file2: str) -> Tuple[int, str]:
    """
    Check the differences between two files
    """
    for f in [file1, file2]:
        if not os.path.isfile(f):
            return (1, f"{f} does not exist\n")

    diff_func = cib_diff if os.path.basename(file1) == constants.CIB_F else txt_diff
    return diff_func(file1, file2)


def get_distro_info() -> str:
    """
    Get distribution information
    """
    res = None
    if os.path.exists(constants.OSRELEASE):
        logger.debug2(f"Using {constants.OSRELEASE} to get distribution info")
        res = re.search("PRETTY_NAME=\"(.*)\"", crmutils.read_from_file(constants.OSRELEASE))
    elif shutil.which("lsb_release"):
        logger.debug2("Using lsb_release to get distribution info")
        out = sh.LocalShell().get_stdout_or_raise_error("lsb_release -d")
        res = re.search(r"Description:\s+(.*)", out)
    return res.group(1) if res else "Unknown"


def dump_logset(context: core.Context, logf: str, create_dir: bool = False) -> None:
    """
    Dump the log set into the specified output file
    """
    logf_set, logf_type = arch_logs(context, logf)
    if not logf_set:
        logger.debug2(f"{logf} is not in timespan {get_timespan_str(context)}")
        return

    out_string = ""

    if logf_type == LogType.IRREGULAR:
        for f in logf_set:
            out_string += print_logseg(f, 0, 0)
    else:
        newest, oldest = logf_set[0], logf_set[-1]
        middle_set = logf_set[1:-1]

        if len(logf_set) == 1:
            out_string += print_logseg(newest, context.from_time, context.to_time)
        else:
            out_string += print_logseg(oldest, context.from_time, 0)
            for f in middle_set:
                out_string += print_logseg(f, 0, 0)
            out_string += print_logseg(newest, 0, context.to_time)

    if out_string:
        if create_dir:
            basename, dirname = os.path.basename(logf), os.path.dirname(logf)
            dest_dir = os.path.join(
                context.work_dir,
                f'{basename}.d',
                dirname.lstrip('/')
            )
            crmutils.mkdirp(dest_dir)
            outf = os.path.join(dest_dir, basename)
        else:
            outf = os.path.join(context.work_dir, os.path.basename(logf))
        crmutils.str2file(out_string.strip('\n'), outf)
        logger.debug(f"Dump {logf} into {real_path(outf)}")


def find_files_in_timespan(context: core.Context, target_dir_list: List[str]) -> List[str]:
    """
    Get a list of files in the target directories with creation time in the timespan
    """
    file_list = []

    for target_dir in target_dir_list:
        if not os.path.isdir(target_dir):
            continue

        for root, dirs, files in os.walk(target_dir):
            for file in files:
                file_path = os.path.join(root, file)
                file_stat = os.stat(file_path)
                if context.from_time <= file_stat.st_ctime <= context.to_time:
                    file_list.append(file_path)

    return file_list


def find_first_timestamp(data: List[str], log_file: str) -> float:
    """
    Find the first timestamp in the given list of log line
    """
    for line in data:
        timestamp = get_timestamp(line, log_file)
        if timestamp:
            return timestamp
    return None


def filter_lines(data: str, from_line: int, to_line: int) -> str:
    """
    Filter lines from the given data based on the specified line range.
    """
    lines = data.split('\n')
    filtered_lines = [
        line + '\n' 
        for count, line in enumerate(lines, start=1) 
        if from_line <= count <= to_line
    ]
    return ''.join(filtered_lines)


def determin_log_format(data: str) -> str:
    """
    Determines the log format based on the given log line
    """
    for line in head(constants.CHECK_LOG_LINES, data):
        _list = line.split()
        if not _list:
            continue
        # syslog format:
        # Feb 12 18:30:08 15sp1-1 kernel: e820: BIOS-provided physical RAM map:
        if len(_list) >= 3 and crmutils.parse_time(' '.join(_list[0:3]), quiet=True):
            return "syslog"
        # rfc5424 format:
        # 2003-10-11T22:14:15.003Z mymachine.example.com su
        if crmutils.parse_time(_list[0], quiet=True):
            return "rfc5424"
        if len(_list) > 1 and crmutils.parse_time(_list[1], quiet=True):
            return "legacy"
    return None


def findln_by_timestamp(data: str, given_timestamp: float, log_file: str) -> int:
    """
    Get line number of the specific time stamp
    """
    data_list = data.split('\n')
    first, last = 1, len(data_list)

    while first <= last:
        middle = (last + first) // 2
        trycnt = 10
        while trycnt > 0:
            middle_timestamp = get_timestamp(data_list[middle - 1], log_file)
            if middle_timestamp:
                break
            # shift the whole first-last segment
            trycnt -= 1
            prevmid = middle
            while prevmid == middle:
                first -= 1
                if first < 1:
                    first = 1
                last -= 1
                if last < first:
                    last = first
                prevmid = middle
                middle = (last + first) // 2
                if first == last:
                    break

        if not middle_timestamp:
            return None
        if middle_timestamp > given_timestamp:
            last = middle - 1
        elif middle_timestamp < given_timestamp:
            first = middle + 1
        else:
            break

    return middle


def get_pkg_mgr() -> str:
    """
    Get the package manager available in the system
    """
    pkg_mgr_candidates = {
        "rpm": "rpm",
        "dpkg": "deb"
    }
    for pkg_mgr, pkg_mgr_name in pkg_mgr_candidates.items():
        if shutil.which(pkg_mgr):
            return pkg_mgr_name

    logger.warning("Unknown package manager!")
    return ""


def get_timestamp_from_time_line(time_line: str, stamp_type: str, log_file: str):
    timestamp = crmutils.parse_to_timestamp(time_line, quiet=True)
    if timestamp and stamp_type == "syslog":
        now = datetime.datetime.now()
        # got a timestamp in the future
        if timestamp > now.timestamp():
            # syslog doesn't have year info, so we need to guess it
            mtime = os.path.getmtime(log_file)
            mtime = datetime.datetime.fromtimestamp(mtime)
            # assume the log is from last year
            if mtime.year == now.year:
                time_line += f" {mtime.year-1}"
            # assume the log is from that year
            elif mtime.year < now.year:
                time_line += f" {mtime.year}"
            # it's impossible that the log is from next year
            else:
                return None
        return crmutils.parse_to_timestamp(time_line, quiet=True)
    else:
        return timestamp


def get_timestamp(line: str, log_file: str) -> float:
    """
    Get timestamp for the given line
    """
    if not line or not constants.STAMP_TYPE:
        return None

    stamp_type = constants.STAMP_TYPE
    if stamp_type == "rfc5424":
        time_line = line.split()[0]
    elif stamp_type == "syslog":
        time_line = ' '.join(line.split()[0:3])
    elif stamp_type == "legacy":
        time_line = line.split()[1]

    return get_timestamp_from_time_line(time_line, stamp_type, log_file)



def head(n: int, indata: str) -> List[str]:
    return indata.split('\n')[:n]


def is_our_log(context: core.Context, logf: str) -> int:
    """
    Check if the log contains a piece of our segment

    Return log type LogType
    """
    data = crmutils.read_from_file(logf)
    if not data:
        return LogType.EMPTY
    stamp_type = determin_log_format(data)
    if not stamp_type:
        return LogType.IRREGULAR
    constants.STAMP_TYPE = stamp_type

    first_time = find_first_timestamp(head(constants.CHECK_LOG_LINES, data), logf)
    last_time = find_first_timestamp(tail(constants.CHECK_LOG_LINES, data), logf)
    from_time = context.from_time
    to_time = context.to_time

    if not first_time or not last_time:
        return LogType.IRREGULAR
    if from_time > last_time:
        return LogType.BEFORE_TIMESPAN
    if from_time >= first_time or to_time >= first_time:
        return LogType.GOOD
    return LogType.AFTER_TIMESPAN


def create_description_template(context: core.Context) -> None:
    """
    Create description template, editing, and other notes
    """
    out_string = constants.DECRIPTION_TMPLATE.format(now(), ' '.join(sys.argv[1:]))

    for n in context.node_list:
        sysinfo_node_f = os.path.join(context.work_dir, n, constants.SYSINFO_F)
        if os.path.isfile(sysinfo_node_f):
            out_string += f"[Info from node {n}]:\n"
            out_string += crmutils.read_from_file(sysinfo_node_f)
            out_string += "\n\n\n\n"

    description_f = os.path.join(context.work_dir, constants.DESCRIPTION_F)
    crmutils.str2file(out_string, description_f)


def print_logseg(log_file: str, from_time: float, to_time: float) -> str:
    """
    Print the log segment specified by the given timestamps
    """
    data = crmutils.read_from_file(log_file)
    if not data:
        return ""

    from_line = 1 if from_time == 0 else findln_by_timestamp(data, from_time, log_file)
    to_line = len(data.split('\n')) if to_time == 0 else findln_by_timestamp(data, to_time, log_file)

    if from_line is None or to_line is None:
        return ""

    logger.debug2("Including segment [%d-%d] from %s", from_line, to_line, log_file)
    return filter_lines(data, from_line, to_line)


def tail(n: int, indata: str) -> List[str]:
    return indata.split('\n')[-n:]


def txt_diff(file1: str, file2: str) -> Tuple[int, str]:
    cmd = f"diff -bBu {file1} {file2}"
    rc, out, _ = ShellUtils().get_stdout_stderr(cmd)
    return rc, out


class Sanitizer:
    """
    A class containing methods for sanitizing sensitive data in CIB and PE files
    """
    DEFAULT_RULE_LIST = ["passw.*"]

    def __init__(self, context: core.Context) -> None:
        self.file_list_in_workdir = []
        self.context = context
        self.cib_data = None
        self.sensitive_regex_set = set()
        self.sensitive_value_list_with_raw_option = []
        self.sensitive_value_list = []
        self.sensitive_key_list = []

    def prepare(self) -> None:
        """
        Prepare the data and files for the sanitization process
        """
        self.cib_data = self._load_cib_from_work_dir()
        if not self.cib_data:
            return False
        self._parse_sensitive_set()
        self._extract_sensitive_value_list()

        if self._include_sensitive_data():
            if not self.context.sanitize:
                logger.warning("Some PE/CIB/log files contain possibly sensitive data")
                logger.warning("Using \"-s\" option can replace sensitive data")
                return False
            self._get_file_list_in_work_dir()
        else:
            self.context.sanitize = False

        return True

    def _include_sensitive_data(self) -> List[str]:
        """
        Check whether contain sensitive data
        """
        return self.sensitive_value_list_with_raw_option or self.sensitive_value_list

    def _get_file_list_in_work_dir(self) -> List[str]:
        """
        Get all files in work directory
        """
        for dirpath, dirnames, filenames in os.walk(self.context.work_dir):
            for _file in filenames:
                self.file_list_in_workdir.append(os.path.join(dirpath, _file))

    def _load_cib_from_work_dir(self) -> None:
        """
        Load CIB data from the working directory
        """
        cib_file_list = glob.glob(f"{self.context.work_dir}/*/{constants.CIB_F}")
        if not cib_file_list:
            return None
        return crmutils.read_from_file(cib_file_list[0])

    def _parse_sensitive_set(self) -> None:
        """
        Parse sensitive regex from -E option and config.report.sanitize_rule
        """
        # from command line option -p
        patt_set = set(self.context.sensitive_regex_list)
        # from /etc/crm/crm.conf
        if config.report.sanitize_rule:
            patt_set |= set(re.split(r'\s*\|\s*|\s+', config.report.sanitize_rule.strip('|')))
        if patt_set:
            self.context.sanitize = True
        # Not set from -p option and crm.conf, use default
        else:
            patt_set = set(Sanitizer.DEFAULT_RULE_LIST)
        logger.debug2(f"Regex set to match sensitive data: {patt_set}")
        self.sensitive_regex_set = patt_set

    def _extract_sensitive_value_list(self) -> None:
        """
        Extract sensitive value list from cib data
        """
        for patt in self.sensitive_regex_set:
            if ':' in patt:
                rule, option = patt.split(':')
                if option == 'raw':
                    self.sensitive_value_list_with_raw_option += self._extract_from_cib(rule)
                else:
                    logger.warning(f"For sanitize pattern {patt}, option should be \"raw\"")
            else:
                self.sensitive_value_list += self._extract_from_cib(patt)
                self.sensitive_key_list.append(patt.strip('.*?')+'.*?')

    def _extract_from_cib(self, rule:str) -> List[str]:
        name_patt = rule.strip('?')+'?'
        value_list = re.findall(f'name="({name_patt})" value="(.*?)"', self.cib_data)
        return [value[1] for value in value_list]

    def _sub_sensitive_string(self, data: str) -> str:
        """
        Do the replacement job

        For the raw sanitize pattern, replace exactly the value
        For the key:value nvpair sanitize pattern, replace the value in which line contain the key
        """
        result = data
        replace_raw_n: int = 0
        replace_n: int = 0

        if self.sensitive_value_list_with_raw_option:
            patt_str = '|'.join([re.escape(patt) for patt in self.sensitive_value_list_with_raw_option])
            result, replace_raw_n = re.subn(r'\b({})\b'.format(patt_str), "******", data)
        if self.sensitive_value_list:
            key_str = '|'.join(self.sensitive_key_list)
            patt_str = '|'.join([re.escape(patt) for patt in self.sensitive_value_list])
            result, replace_n = re.subn(f'({key_str})({patt_str})', '\\1******', result)

        return "" if (replace_raw_n == 0 and replace_n == 0) else result


    def sanitize(self) -> None:
        """
        Replace and overwrite files containing sensitive data
        """
        if not self.context.sanitize:
            return
        for f in self.file_list_in_workdir:
            data = crmutils.read_from_file(f)
            if not data:
                continue
            replaced_str = self._sub_sensitive_string(data)
            if replaced_str:
                logger.debug("Replace sensitive info for %s", f)
                write_to_file(replaced_str, f)


def do_sanitize(context: core.Context) -> None:
    """
    Perform sanitization by replacing sensitive information in CIB/PE/other logs data with '*'
    """
    inst = Sanitizer(context)
    if inst.prepare():
        inst.sanitize()


class Package:
    """
    A class to retrieve package versions and verify packages
    on various distros
    """
    def __init__(self, packages: str) -> None:
        self.pkg_type = get_pkg_mgr()
        self.packages = packages

    def pkg_ver_deb(self) -> str:
        cmd = f"dpkg-query -W -f='${{Package}}    ${{Version}}.${{Architecture}}\n' {self.packages}"
        _, out, _ = ShellUtils().get_stdout_stderr(cmd)
        return '\n'.join([line for line in out.splitlines() if "no packages found" not in line])

    def pkg_ver_rpm(self) -> str:
        _, out, _ = ShellUtils().get_stdout_stderr(f"rpm -q {self.packages}")
        return '\n'.join([line for line in out.splitlines() if "not installed" not in line])

    def version(self) -> str:
        if not self.pkg_type:
            return ""
        return getattr(self, f"pkg_ver_{self.pkg_type}")()

    def verify_deb(self) -> str:
        cmd = f"dpkg --verify {self.packages}"
        _, out, _ = ShellUtils().get_stdout_stderr(cmd)
        return '\n'.join([line for line in out.splitlines() if "not installed" not in line])

    def verify_rpm(self) -> str:
        cmd = f"rpm --verify {self.packages}"
        _, out, _ = ShellUtils().get_stdout_stderr(cmd)
        return '\n'.join([line for line in out.splitlines() if "not installed" not in line])

    def verify(self) -> str:
        if not self.pkg_type:
            return ""
        return getattr(self, f"verify_{self.pkg_type}")()


def write_to_file(data: str, tofile: str) -> None:
    _open = crmutils.get_open_method(tofile)
    with _open(tofile, 'w') as f:
        if _open == open:
            f.write(data)
        else:
            f.write(data.encode('utf-8'))


def parse_to_timestamp(time: str) -> Optional[float]:
    """
    Parses the input time string and converts it to a timestamp
    """
    time_format_mapping = {
        'Y': 365,  # 1 year is approximately 365 days
        'm': 30,   # 1 month is approximately 30 days
        'd': 1,
        'H': 1 / 24,  # 1 hour is 1/24 of a day
        'M': 1 / 1440  # 1 minute is 1/1440 of a day
    }

    # Match the input time string to the format
    match = re.match(r'^-?([1-9][0-9]*)([YmdHM])$', time)

    if not match:
        res = crmutils.parse_to_timestamp(time, quiet=True)
        if res:
            return res
        logger.error(f"Invalid time string '{time}'")
        logger.error('Try these formats like: 2pm; "2019/9/5 12:30"; "09-Sep-07 2:00"; "[1-9][0-9]*[YmdHM]"')
        raise ReportGenericError

    number_str, flag = match.groups()
    number = int(number_str) * time_format_mapping[flag]
    timedelta = datetime.timedelta(days=number)

    # Calculate the timestamp
    timestamp = (datetime.datetime.now() - timedelta).timestamp()

    return timestamp


def ts_to_str(timestamp: float) -> str:
    """
    Convert timestamp to date string
    """
    return dt_to_str(ts_to_dt(timestamp))


def ts_to_dt(timestamp: float) -> datetime.datetime:
    """
    Convert timestamp to datetime.datetime object, consider utc offset
    """
    dt = crmutils.timestamp_to_datetime(timestamp)
    dt += tz.tzlocal().utcoffset(dt)
    return dt


def dt_to_str(dt: datetime.datetime, form: str = constants.TIME_FORMAT) -> str:
    return dt.strftime(form)


def now(form: str = constants.TIME_FORMAT) -> str:
    return dt_to_str(datetime.datetime.now(), form=form)


def get_cmd_output(cmd: str, timeout: int = None) -> str:
    """
    Get the output of a command, include stdout and stderr
    """
    out_str = ""
    _, out, err = ShellUtils().get_stdout_stderr(cmd, timeout=timeout)
    if out:
        out_str += f"{out}\n"
    if err:
        out_str += f"{err}\n"
    return crmutils.strip_ansi_escape_sequences(out_str)


def get_timespan_str(context: core.Context) -> str:
    from_time_str = ts_to_str(context.from_time)
    to_time_str = ts_to_str(context.to_time)
    return f"{from_time_str} - {to_time_str}"


def print_traceback():
    traceback.print_exc()
    sys.stdout.flush()


def real_path(target_file: str) -> str:
    return '/'.join(target_file.split('/')[3:])


def mark_duplicate_basenames(file_path_list: list[str]) -> list[tuple[str, bool]]:
    """
    Mark the paths with the same basename
    Return a list of tuples, each tuple contains the path and a boolean value
    indicating whether the path has the same basename with others
    """
    counter = Counter([os.path.basename(path) for path in file_path_list])
    return [(path, counter[os.path.basename(path)] > 1) for path in file_path_list]
# vim:ts=4:sw=4:et:



from __future__ import print_function
from __future__ import unicode_literals
# Copyright (C) 2014 Kristoffer Gronlund <kgronlund@suse.com>
# See COPYING for license information.
try:
    from unittest import mock
except ImportError:
    import mock

from crmsh import cibconfig
from lxml import etree
from crmsh import xmlutil, utils

factory = cibconfig.cib_factory


def setup_function():
    "set up test fixtures"
    from crmsh import idmgmt
    idmgmt.clear()
    factory._push_state()


def teardown_function():
    factory._pop_state()


def test_bug41660_1():
    xml = """<primitive id="bug41660" class="ocf" provider="pacemaker" type="Dummy"> \
    <meta_attributes id="bug41660-meta"> \
    <nvpair id="bug41660-meta-target-role" name="target-role" value="Stopped"/> \
    </meta_attributes> \
    </primitive>
"""
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    print(etree.tostring(obj.node))
    data = obj.repr_cli(format_mode=-1)
    print(data)
    exp = 'primitive bug41660 ocf:pacemaker:Dummy meta target-role=Stopped'
    assert data == exp
    assert obj.cli_use_validate()

    commit_holder = factory.commit
    try:
        factory.commit = lambda *args: True
        from crmsh.ui_resource import set_deep_meta_attr
        set_deep_meta_attr("bug41660", "target-role", "Started")
        assert ['Started'] == obj.node.xpath('.//nvpair[@name="target-role"]/@value')
    finally:
        factory.commit = commit_holder


def test_bug41660_2():
    xml = """
<clone id="libvirtd-clone">
 <primitive class="lsb" id="libvirtd" type="libvirtd">
  <operations>
   <op id="libvirtd-monitor-interval-15" interval="15" name="monitor" start-delay="15" timeout="15"/>
   <op id="libvirtd-start-interval-0" interval="0" name="start" on-fail="restart" timeout="15"/>
   <op id="libvirtd-stop-interval-0" interval="0" name="stop" on-fail="ignore" timeout="15"/>
  </operations>
  <meta_attributes id="libvirtd-meta_attributes"/>
 </primitive>
 <meta_attributes id="libvirtd-clone-meta">
  <nvpair id="libvirtd-interleave" name="interleave" value="true"/>
  <nvpair id="libvirtd-ordered" name="ordered" value="true"/>
  <nvpair id="libvirtd-clone-meta-target-role" name="target-role" value="Stopped"/>
 </meta_attributes>
</clone>
"""
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    #data = obj.repr_cli(format_mode=-1)
    #print data
    #exp = 'clone libvirtd-clone libvirtd meta interleave=true ordered=true target-role=Stopped'
    #assert data == exp
    #assert obj.cli_use_validate()

    print(etree.tostring(obj.node))

    commit_holder = factory.commit
    try:
        factory.commit = lambda *args: True
        from crmsh.ui_resource import set_deep_meta_attr
        print("PRE", etree.tostring(obj.node))
        set_deep_meta_attr("libvirtd-clone", "target-role", "Started")
        print("POST", etree.tostring(obj.node))
        assert ['Started'] == obj.node.xpath('.//nvpair[@name="target-role"]/@value')
    finally:
        factory.commit = commit_holder


def test_bug41660_3():
    xml = """
<clone id="libvirtd-clone">
 <primitive class="lsb" id="libvirtd" type="libvirtd">
  <operations>
   <op id="libvirtd-monitor-interval-15" interval="15" name="monitor" start-delay="15" timeout="15"/>
   <op id="libvirtd-start-interval-0" interval="0" name="start" on-fail="restart" timeout="15"/>
   <op id="libvirtd-stop-interval-0" interval="0" name="stop" on-fail="ignore" timeout="15"/>
  </operations>
  <meta_attributes id="libvirtd-meta_attributes"/>
 </primitive>
 <meta_attributes id="libvirtd-clone-meta_attributes">
 <nvpair id="libvirtd-clone-meta_attributes-target-role" name="target-role" value="Stopped"/>
 </meta_attributes>
</clone>
"""
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print(data)
    exp = 'clone libvirtd-clone libvirtd meta target-role=Stopped'
    assert data == exp
    assert obj.cli_use_validate()

    commit_holder = factory.commit
    try:
        factory.commit = lambda *args: True
        from crmsh.ui_resource import set_deep_meta_attr
        set_deep_meta_attr("libvirtd-clone", "target-role", "Started")
        assert ['Started'] == obj.node.xpath('.//nvpair[@name="target-role"]/@value')
    finally:
        factory.commit = commit_holder


def test_comments():
    xml = """<cib epoch="25" num_updates="1" admin_epoch="0" validate-with="pacemaker-1.2" cib-last-written="Thu Mar  6 15:53:49 2014" update-origin="beta1" update-client="cibadmin" update-user="root" crm_feature_set="3.0.8" have-quorum="1" dc-uuid="1">
  <configuration>
    <crm_config>
      <cluster_property_set id="cib-bootstrap-options">
        <nvpair id="cib-bootstrap-options-dc-version" name="dc-version" value="1.1.11-3.3-3ca8c3b"/>
        <nvpair id="cib-bootstrap-options-cluster-infrastructure" name="cluster-infrastructure" value="corosync"/>
        <!--# COMMENT TEXT 1 -->
      </cluster_property_set>
    </crm_config>
    <nodes>
      <node uname="beta1" id="1">
        <!--# COMMENT TEXT 2 -->
      </node>
    </nodes>
    <resources/>
    <constraints/>
    <rsc_defaults>
      <meta_attributes id="rsc-options">
        <nvpair name="resource-stickiness" value="1" id="rsc-options-resource-stickiness"/>
        <!--# COMMENT TEXT 3 -->
      </meta_attributes>
    </rsc_defaults>
  </configuration>
  <status>
    <node_state id="1" uname="beta1" in_ccm="true" crmd="online" crm-debug-origin="do_state_transition" join="member" expected="member">
      <lrm id="1">
        <lrm_resources/>
      </lrm>
      <transient_attributes id="1">
        <instance_attributes id="status-1">
          <nvpair id="status-1-shutdown" name="shutdown" value="0"/>
          <nvpair id="status-1-probe_complete" name="probe_complete" value="true"/>
        </instance_attributes>
      </transient_attributes>
    </node_state>
  </status>
</cib>"""
    elems = etree.fromstring(xml)
    xmlutil.sanitize_cib(elems)
    assert xmlutil.xml_tostring(elems).count("COMMENT TEXT") == 3


def test_eq1():
    xml1 = """<cluster_property_set id="cib-bootstrap-options">
    <nvpair id="cib-bootstrap-options-stonith-enabled" name="stonith-enabled" value="true"></nvpair>
    <nvpair id="cib-bootstrap-options-stonith-timeout" name="stonith-timeout" value="180"></nvpair>
    <nvpair id="cib-bootstrap-options-symmetric-cluster" name="symmetric-cluster" value="false"></nvpair>
    <nvpair id="cib-bootstrap-options-no-quorum-policy" name="no-quorum-policy" value="freeze"></nvpair>
    <nvpair id="cib-bootstrap-options-batch-limit" name="batch-limit" value="20"></nvpair>
    <nvpair id="cib-bootstrap-options-dc-version" name="dc-version" value="1.1.10-c1a326d"></nvpair>
    <nvpair id="cib-bootstrap-options-cluster-infrastructure" name="cluster-infrastructure" value="corosync"></nvpair>
    <nvpair id="cib-bootstrap-options-last-lrm-refresh" name="last-lrm-refresh" value="1391433789"></nvpair>
    <nvpair id="cib-bootstrap-options-is-managed-default" name="is-managed-default" value="true"></nvpair>
  </cluster_property_set>
  """
    xml2 = """<cluster_property_set id="cib-bootstrap-options">
    <nvpair id="cib-bootstrap-options-stonith-enabled" name="stonith-enabled" value="true"></nvpair>
    <nvpair id="cib-bootstrap-options-stonith-timeout" name="stonith-timeout" value="180"></nvpair>
    <nvpair id="cib-bootstrap-options-symmetric-cluster" name="symmetric-cluster" value="false"></nvpair>
    <nvpair id="cib-bootstrap-options-no-quorum-policy" name="no-quorum-policy" value="freeze"></nvpair>
    <nvpair id="cib-bootstrap-options-batch-limit" name="batch-limit" value="20"></nvpair>
    <nvpair id="cib-bootstrap-options-dc-version" name="dc-version" value="1.1.10-c1a326d"></nvpair>
    <nvpair id="cib-bootstrap-options-cluster-infrastructure" name="cluster-infrastructure" value="corosync"></nvpair>
    <nvpair id="cib-bootstrap-options-last-lrm-refresh" name="last-lrm-refresh" value="1391433789"></nvpair>
    <nvpair id="cib-bootstrap-options-is-managed-default" name="is-managed-default" value="true"></nvpair>
  </cluster_property_set>
    """
    e1 = etree.fromstring(xml1)
    e2 = etree.fromstring(xml2)
    assert xmlutil.xml_equals(e1, e2, show=True)


def test_pcs_interop_1():
    """
    pcs<>crmsh interop bug
    """

    xml = """<clone id="dummies">
        <meta_attributes id="dummies-meta">
          <nvpair name="globally-unique" value="false" id="dummies-meta-globally-unique"/>
        </meta_attributes>
        <meta_attributes id="dummies-meta_attributes">
          <nvpair id="dummies-meta_attributes-target-role" name="target-role" value="Stopped"/>
        </meta_attributes>
        <primitive id="dummy-1" class="ocf" provider="heartbeat" type="Dummy"/>
      </clone>"""
    elem = etree.fromstring(xml)
    from crmsh.ui_resource import set_deep_meta_attr_node

    assert len(elem.xpath(".//meta_attributes/nvpair[@name='target-role']")) == 1

    print("BEFORE:", etree.tostring(elem))

    set_deep_meta_attr_node(elem, 'target-role', 'Stopped')

    print("AFTER:", etree.tostring(elem))

    assert len(elem.xpath(".//meta_attributes/nvpair[@name='target-role']")) == 1


def test_bnc878128():
    """
    L3: "crm configure show" displays XML information instead of typical crm output.
    """
    xml = """<rsc_location id="cli-prefer-dummy-resource" rsc="dummy-resource"
role="Started">
  <rule id="cli-prefer-rule-dummy-resource" score="INFINITY">
    <expression id="cli-prefer-expr-dummy-resource" attribute="#uname"
operation="eq" value="x64-4"/>
    <date_expression id="cli-prefer-lifetime-end-dummy-resource" operation="lt"
end="2014-05-17 17:56:11Z"/>
  </rule>
</rsc_location>"""
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print("OUTPUT:", data)
    exp = 'location cli-prefer-dummy-resource dummy-resource role=Started rule #uname eq x64-4 and date lt "2014-05-17 17:56:11Z"'
    assert data == exp
    assert obj.cli_use_validate()


def test_order_without_score_kind():
    """
    Spec says order doesn't require score or kind to be set
    """
    xml = '<rsc_order first="a" first-action="promote" id="order-a-b" then="b" then-action="start"/>'
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print("OUTPUT:", data)
    exp = 'order order-a-b a:promote b:start'
    assert data == exp
    assert obj.cli_use_validate()



def test_bnc878112():
    """
    crm configure group can hijack a cloned primitive (and then crash)
    """
    obj1 = factory.create_object('primitive', 'p1', 'Dummy')
    assert obj1 is True
    obj2 = factory.create_object('group', 'g1', 'p1')
    assert obj2 is True
    obj3 = factory.create_object('group', 'g2', 'p1')
    print(obj3)
    assert obj3 is False


def test_copy_nvpairs():
    from crmsh.cibconfig import copy_nvpairs

    to = etree.fromstring('''
    <node>
    <nvpair name="stonith-enabled" value="true"/>
    </node>
    ''')
    copy_nvpairs(to, etree.fromstring('''
    <node>
    <nvpair name="stonith-enabled" value="false"/>
    </node>
    '''))

    assert ['stonith-enabled'] == to.xpath('./nvpair/@name')
    assert ['false'] == to.xpath('./nvpair/@value')

    copy_nvpairs(to, etree.fromstring('''
    <node>
    <nvpair name="stonith-enabled" value="true"/>
    </node>
    '''))

    assert ['stonith-enabled'] == to.xpath('./nvpair/@name')
    assert ['true'] == to.xpath('./nvpair/@value')


def test_pengine_test():
    xml = '''<primitive class="ocf" id="rsc1" provider="pacemaker" type="Dummy">
        <instance_attributes id="rsc1-instance_attributes-1">
          <nvpair id="rsc1-instance_attributes-1-state" name="state" value="/var/run/Dummy-rsc1-clusterA"/>
          <rule id="rsc1-instance_attributes-1-rule-1" score="0">
            <expression id="rsc1-instance_attributes-1-rule-1-expr-1" attribute="#cluster-name" operation="eq" value="clusterA"/>
          </rule>
        </instance_attributes>
        <instance_attributes id="rsc1-instance_attributes-2">
          <nvpair id="rsc1-instance_attributes-2-state" name="state" value="/var/run/Dummy-rsc1-clusterB"/>
          <rule id="rsc1-instance_attributes-2-rule-1" score="0">
            <expression id="rsc1-instance_attributes-2-rule-1-expr-1" attribute="#cluster-name" operation="eq" value="clusterB"/>
          </rule>
        </instance_attributes>
        <operations>
          <op id="rsc1-monitor-10s" interval="10s" name="monitor"/>
        </operations>
      </primitive>'''
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print("OUTPUT:", data)
    exp = 'primitive rsc1 ocf:pacemaker:Dummy params rule 0: #cluster-name eq clusterA state="/var/run/Dummy-rsc1-clusterA" params rule 0: #cluster-name eq clusterB state="/var/run/Dummy-rsc1-clusterB" op monitor interval=10s'
    assert data == exp
    assert obj.cli_use_validate()


def test_tagset():
    xml = '''<primitive class="ocf" id="%s" provider="pacemaker" type="Dummy"/>'''
    tag = '''<tag id="t0"><obj_ref id="r1"/><obj_ref id="r2"/></tag>'''
    factory.create_from_node(etree.fromstring(xml % ('r1')))
    factory.create_from_node(etree.fromstring(xml % ('r2')))
    factory.create_from_node(etree.fromstring(xml % ('r3')))
    factory.create_from_node(etree.fromstring(tag))
    elems = factory.get_elems_on_tag("tag:t0")
    assert set(x.obj_id for x in elems) == set(['r1', 'r2'])


def test_op_role():
    xml = '''<primitive class="ocf" id="rsc2" provider="pacemaker" type="Dummy">
        <operations>
          <op id="rsc2-monitor-10s" interval="10s" name="monitor" role="Stopped"/>
        </operations>
      </primitive>'''
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print("OUTPUT:", data)
    exp = 'primitive rsc2 ocf:pacemaker:Dummy op monitor interval=10s role=Stopped'
    assert data == exp
    assert obj.cli_use_validate()


def test_nvpair_no_value():
    xml = '''<primitive class="ocf" id="rsc3" provider="heartbeat" type="Dummy">
        <instance_attributes id="rsc3-instance_attributes-1">
          <nvpair id="rsc3-instance_attributes-1-verbose" name="verbose"/>
          <nvpair id="rsc3-instance_attributes-1-verbase" name="verbase" value=""/>
          <nvpair id="rsc3-instance_attributes-1-verbese" name="verbese" value=" "/>
        </instance_attributes>
      </primitive>'''
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print("OUTPUT:", data)
    exp = 'primitive rsc3 Dummy params verbose verbase="" verbese=" "'
    assert data == exp
    assert obj.cli_use_validate()


def test_delete_ticket():
    xml0 = '<primitive id="daa0" class="ocf" provider="heartbeat" type="Dummy"/>'
    xml1 = '<primitive id="daa1" class="ocf" provider="heartbeat" type="Dummy"/>'
    xml2 = '''<rsc_ticket id="taa0" ticket="taaA">
        <resource_set id="taa0-0">
          <resource_ref id="daa0"/>
          <resource_ref id="daa1"/>
        </resource_set>
      </rsc_ticket>'''
    for x in (xml0, xml1, xml2):
        data = etree.fromstring(x)
        obj = factory.create_from_node(data)
        assert obj is not None
        data = obj.repr_cli(format_mode=-1)

    factory.delete('daa0')
    assert factory.find_object('daa0') is None
    assert factory.find_object('taa0') is not None


def test_quotes():
    """
    Parsing escaped quotes
    """
    xml = '''<primitive class="ocf" id="q1" provider="pacemaker" type="Dummy">
        <instance_attributes id="q1-instance_attributes-1">
          <nvpair id="q1-instance_attributes-1-state" name="state" value="foo&quot;foo&quot;"/>
        </instance_attributes>
    </primitive>
    '''
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    print("OUTPUT:", data)
    exp = 'primitive q1 ocf:pacemaker:Dummy params state="foo\\"foo\\""'
    assert data == exp
    assert obj.cli_use_validate()


def test_nodeattrs():
    """
    bug with parsing node attrs
    """
    xml = '''<node id="1" uname="dell71"> \
  <instance_attributes id="dell71-instance_attributes"> \
    <nvpair name="staging-0-0-placement" value="true" id="dell71-instance_attributes-staging-0-0-placement"/> \
    <nvpair name="meta-0-0-placement" value="true" id="dell71-instance_attributes-meta-0-0-placement"/> \
  </instance_attributes> \
  <instance_attributes id="nodes-1"> \
    <nvpair id="nodes-1-standby" name="standby" value="off"/> \
  </instance_attributes> \
</node>'''

    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    exp = 'node 1: dell71 attributes staging-0-0-placement=true meta-0-0-placement=true attributes standby=off'
    assert data == exp
    assert obj.cli_use_validate()


def test_nodeattrs2():
    xml = """<node id="h04" uname="h04"> \
 <utilization id="h04-utilization"> \
   <nvpair id="h04-utilization-utl_ram" name="utl_ram" value="1200"/> \
   <nvpair id="h04-utilization-utl_cpu" name="utl_cpu" value="200"/> \
 </utilization> \
 <instance_attributes id="nodes-h04"> \
   <nvpair id="nodes-h04-standby" name="standby" value="off"/> \
 </instance_attributes> \
</node>"""
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    assert obj is not None
    data = obj.repr_cli(format_mode=-1)
    exp = 'node h04 utilization utl_ram=1200 utl_cpu=200 attributes standby=off'
    assert data == exp
    assert obj.cli_use_validate()


def test_group_constraint_location():
    """
    configuring a location constraint on a grouped resource is OK
    """
    factory.create_object('node', 'node1')
    factory.create_object('primitive', 'p1', 'Dummy')
    factory.create_object('primitive', 'p2', 'Dummy')
    factory.create_object('group', 'g1', 'p1', 'p2')
    factory.create_object('location', 'loc-p1', 'p1', 'inf:', 'node1')
    c = factory.find_object('loc-p1')
    assert c and c.check_sanity() == utils.VerifyResult.SUCCESS


def test_group_constraint_colocation():
    """
    configuring a colocation constraint on a grouped resource is bad
    """
    factory.create_object('primitive', 'p1', 'Dummy')
    factory.create_object('primitive', 'p2', 'Dummy')
    factory.create_object('group', 'g1', 'p1', 'p2')
    factory.create_object('colocation', 'coloc-p1-p2', 'inf:', 'p1', 'p2')
    c = factory.find_object('coloc-p1-p2')
    rc = c.check_sanity()
    assert c and bool(rc) is True and utils.VerifyResult.WARNING in rc


def test_group_constraint_colocation_rscset():
    """
    configuring a constraint on a grouped resource is bad
    """
    factory.create_object('primitive', 'p1', 'Dummy')
    factory.create_object('primitive', 'p2', 'Dummy')
    factory.create_object('primitive', 'p3', 'Dummy')
    factory.create_object('group', 'g1', 'p1', 'p2')
    factory.create_object('colocation', 'coloc-p1-p2-p3', 'inf:', 'p1', 'p2', 'p3')
    c = factory.find_object('coloc-p1-p2-p3')
    rc = c.check_sanity()
    assert c and bool(rc) is True and utils.VerifyResult.WARNING in rc


def test_clone_constraint_colocation_rscset():
    """
    configuring a constraint on a cloned resource is bad
    """
    factory.create_object('primitive', 'p1', 'Dummy')
    factory.create_object('primitive', 'p2', 'Dummy')
    factory.create_object('primitive', 'p3', 'Dummy')
    factory.create_object('clone', 'c1', 'p1')
    factory.create_object('colocation', 'coloc-p1-p2-p3', 'inf:', 'p1', 'p2', 'p3')
    c = factory.find_object('coloc-p1-p2-p3')
    rc = c.check_sanity()
    assert c and bool(rc) is True and utils.VerifyResult.WARNING in rc


def test_existing_node_resource():
    factory.create_object('primitive', 'ha-one', 'Dummy')

    n = factory.find_node('ha-one')
    assert factory.test_element(n)

    r = factory.find_resource('ha-one')
    assert factory.test_element(r)

    assert n != r

    assert factory.check_structure()
    factory.cli_use_validate_all()

    ok, s = factory.mkobj_set('ha-one')
    assert ok


@mock.patch("crmsh.log.LoggerUtils.line_number")
@mock.patch("crmsh.log.LoggerUtils.incr_lineno")
def test_existing_node_resource_2(mock_incr, mock_line_num):
    obj = cibconfig.mkset_obj()
    assert obj is not None

    from crmsh import clidisplay
    with clidisplay.nopretty():
        text = obj.repr()
    text += "\nprimitive ha-one Dummy"
    ok = obj.save(text)
    assert ok

    obj = cibconfig.mkset_obj()
    assert obj is not None
    with clidisplay.nopretty():
        text2 = obj.repr()

    assert sorted(text.split('\n')) == sorted(text2.split('\n'))


@mock.patch("crmsh.log.LoggerUtils.line_number")
@mock.patch("crmsh.log.LoggerUtils.incr_lineno")
def test_id_collision_breakage_1(mock_incr, mock_line_num):
    from crmsh import clidisplay

    obj = cibconfig.mkset_obj()
    assert obj is not None
    with clidisplay.nopretty():
        original_cib = obj.repr()
    print(original_cib)

    obj = cibconfig.mkset_obj()
    assert obj is not None

    ok = obj.save("""node node1
primitive p0 ocf:pacemaker:Dummy
primitive p1 ocf:pacemaker:Dummy
primitive p2 ocf:heartbeat:Delay \
    params startdelay=2 mondelay=2 stopdelay=2
primitive p3 ocf:pacemaker:Dummy
clone c1 p1
op_defaults timeout=60s
""")
    assert ok

    obj = cibconfig.mkset_obj()
    assert obj is not None
    ok = obj.save("""op_defaults timeout=2m
node node1 \
    attributes mem=16G
primitive p1 ocf:heartbeat:Dummy \
    op monitor interval=60m \
    op monitor interval=120m OCF_CHECK_LEVEL=10
""")
    assert ok

    obj = cibconfig.mkset_obj()
    with clidisplay.nopretty():
        text = obj.repr()
    text = text + "\nprimitive p2 ocf:heartbeat:Dummy"
    ok = obj.save(text)
    assert ok

    obj = cibconfig.mkset_obj()
    with clidisplay.nopretty():
        text = obj.repr()
    text = text + "\ngroup g1 p1 p2"
    ok = obj.save(text)
    assert ok

    obj = cibconfig.mkset_obj("g1")
    with clidisplay.nopretty():
        text = obj.repr()
    text = text.replace("group g1 p1 p2", "group g1 p1 p3")
    text = text + "\nprimitive p3 ocf:heartbeat:Dummy"
    ok = obj.save(text)
    assert ok

    obj = cibconfig.mkset_obj("g1")
    with clidisplay.nopretty():
        print(obj.repr().strip())
        assert obj.repr().strip() == "group g1 p1 p3"

    obj = cibconfig.mkset_obj()
    assert obj is not None
    ok = obj.save(original_cib)
    assert ok
    obj = cibconfig.mkset_obj()
    with clidisplay.nopretty():
        print("*** ORIGINAL")
        print(original_cib)
        print("*** NOW")
        print(obj.repr())
        assert original_cib == obj.repr()


@mock.patch("crmsh.log.LoggerUtils.line_number")
@mock.patch("crmsh.log.LoggerUtils.incr_lineno")
def test_id_collision_breakage_3(mock_incr, mock_line_num):
    from crmsh import clidisplay

    obj = cibconfig.mkset_obj()
    assert obj is not None
    with clidisplay.nopretty():
        original_cib = obj.repr()
    print(original_cib)

    obj = cibconfig.mkset_obj()
    assert obj is not None
    ok = obj.save("""node node1
primitive node1 Dummy params fake=something
    """)
    assert ok

    print("** baseline")
    obj = cibconfig.mkset_obj()
    assert obj is not None
    with clidisplay.nopretty():
        print(obj.repr())

    obj = cibconfig.mkset_obj()
    assert obj is not None
    ok = obj.save("""primitive node1 Dummy params fake=something-else
    """, remove=False, method='update')
    assert ok

    print("** end")

    obj = cibconfig.mkset_obj()
    assert obj is not None
    ok = obj.save(original_cib, remove=True, method='replace')
    assert ok
    obj = cibconfig.mkset_obj()
    with clidisplay.nopretty():
        print("*** ORIGINAL")
        print(original_cib)
        print("*** NOW")
        print(obj.repr())
        assert original_cib == obj.repr()


@mock.patch("crmsh.log.LoggerUtils.line_number")
@mock.patch("crmsh.log.LoggerUtils.incr_lineno")
def test_id_collision_breakage_2(mock_incr, mock_line_num):
    from crmsh import clidisplay

    obj = cibconfig.mkset_obj()
    assert obj is not None
    with clidisplay.nopretty():
        original_cib = obj.repr()
    print(original_cib)

    obj = cibconfig.mkset_obj()
    assert obj is not None

    ok = obj.save("""node 168633610: webui
node 168633611: node1
rsc_template web-server apache \
	params port=8000 \
	op monitor interval=10s
primitive d0 Dummy \
	meta target-role=Started
primitive d1 Dummy
primitive d2 Dummy
# Never use this STONITH agent in production!
primitive proxy systemd:haproxy \
	op monitor interval=10s
primitive proxy-vip IPaddr2 \
	params ip=10.13.37.20
primitive srv1 @web-server
primitive srv2 @web-server
primitive vip1 IPaddr2 \
	params ip=10.13.37.21 \
	op monitor interval=20s
primitive vip2 IPaddr2 \
	params ip=10.13.37.22 \
	op monitor interval=20s
primitive virtual-ip IPaddr2 \
	params ip=10.13.37.77 lvs_support=false \
	op start timeout=20 interval=0 \
	op stop timeout=20 interval=0 \
	op monitor interval=10 timeout=20
primitive yet-another-virtual-ip IPaddr2 \
	params ip=10.13.37.72 cidr_netmask=24 \
	op start interval=0 timeout=20 \
	op stop interval=0 timeout=20 \
	op monitor interval=10 timeout=20 \
	meta target-role=Started
group dovip d0 virtual-ip \
	meta target-role=Stopped
group g-proxy proxy-vip proxy
group g-serv1 vip1 srv1
group g-serv2 vip2 srv2
clone d2-clone d2 \
	meta target-role=Started
tag dummytag d0 d1 d1-on-node1 d2 d2-clone
# Never put the two web servers on the same node
colocation co-serv -inf: g-serv1 g-serv2
location d1-on-node1 d1 inf: node1
# Never put any web server or haproxy on webui
location l-avoid-webui { g-proxy g-serv1 g-serv2 } -inf: webui
# Prever to spread groups across nodes
location l-proxy g-proxy 200: node1
location l-serv1 g-serv1 200: node2
location l-serv2 g-serv2 200: node3
property cib-bootstrap-options: \
	have-watchdog=false \
	dc-version="1.1.13+git20150917.20c2178-224.2-1.1.13+git20150917.20c2178" \
	cluster-infrastructure=corosync \
	cluster-name=hacluster \
	stonith-enabled=true \
	no-quorum-policy=ignore
rsc_defaults rsc-options: \
	resource-stickiness=1 \
	migration-threshold=3
op_defaults op-options: \
	timeout=600 \
	record-pending=true
""")
    assert ok

    obj = cibconfig.mkset_obj()
    assert obj is not None
    ok = obj.save(original_cib)
    assert ok
    obj = cibconfig.mkset_obj()
    with clidisplay.nopretty():
        print("*** ORIGINAL")
        print(original_cib)
        print("*** NOW")
        print(obj.repr())
        assert original_cib == obj.repr()


def test_bug_110():
    """
    configuring attribute-based fencing-topology
    """
    factory.create_object(*"primitive stonith-libvirt stonith:fence_sbd".split())
    factory.create_object(*"primitive fence-nova stonith:fence_sbd".split())
    cmd = "fencing_topology attr:OpenStack-role=compute stonith-libvirt,fence-nova".split()
    ok = factory.create_object(*cmd)
    assert ok
    obj = cibconfig.mkset_obj()
    assert obj is not None

    for o in obj.obj_set:
        if o.node.tag == 'fencing-topology':
            assert o.check_sanity() == utils.VerifyResult.SUCCESS


@mock.patch("crmsh.log.LoggerUtils.line_number")
@mock.patch("crmsh.log.LoggerUtils.incr_lineno")
def test_reordering_resource_sets(mock_incr, mock_line_num):
    """
    Can we reorder resource sets?
    """
    from crmsh import clidisplay
    obj1 = factory.create_object('primitive', 'p1', 'Dummy')
    assert obj1 is True
    obj2 = factory.create_object('primitive', 'p2', 'Dummy')
    assert obj2 is True
    obj3 = factory.create_object('primitive', 'p3', 'Dummy')
    assert obj3 is True
    obj4 = factory.create_object('primitive', 'p4', 'Dummy')
    assert obj4 is True
    o1 = factory.create_object('order', 'o1', 'p1', 'p2', 'p3', 'p4')
    assert o1 is True

    obj = cibconfig.mkset_obj('o1')
    assert obj is not None
    rc = obj.save('order o1 p4 p3 p2 p1')
    assert rc == True

    obj2 = cibconfig.mkset_obj('o1')
    with clidisplay.nopretty():
        assert "order o1 p4 p3 p2 p1" == obj2.repr().strip()


def test_bug959895():
    """
    Allow importing XML with cloned groups
    """
    xml = """<clone id="c-bug959895">
    <group id="g-bug959895">
    <primitive id="p-bug959895-a" class="ocf" provider="pacemaker" type="Dummy" />
    <primitive id="p-bug959895-b" class="ocf" provider="pacemaker" type="Dummy" />
    </group>
</clone>
"""
    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    print(etree.tostring(obj.node))
    data = obj.repr_cli(format_mode=-1)
    print(data)
    exp = 'clone c-bug959895 g-bug959895'
    assert data == exp
    assert obj.cli_use_validate()

    commit_holder = factory.commit
    try:
        factory.commit = lambda *args: True
        from crmsh.ui_resource import set_deep_meta_attr
        set_deep_meta_attr("c-bug959895", "target-role", "Started")
        assert ['Started'] == obj.node.xpath('.//nvpair[@name="target-role"]/@value')
    finally:
        factory.commit = commit_holder


def test_node_util_attr():
    """
    Handle node with utitilization before attributes correctly
    """
    xml = """<node id="aberfeldy" uname="aberfeldy">
  <utilization id="nodes-aberfeldy-utilization">
    <nvpair id="nodes-aberfeldy-utilization-cpu" name="cpu" value="2"/>
    <nvpair id="nodes-aberfeldy-utilization-memory" name="memory" value="500"/>
  </utilization>
  <instance_attributes id="nodes-aberfeldy">
    <nvpair id="nodes-aberfeldy-standby" name="standby" value="on"/>
  </instance_attributes>
</node>"""

    data = etree.fromstring(xml)
    obj = factory.create_from_node(data)
    print(etree.tostring(obj.node))
    data = obj.repr_cli(format_mode=-1)
    print(data)
    exp = 'node aberfeldy utilization cpu=2 memory=500 attributes standby=on'
    assert data == exp
    assert obj.cli_use_validate()


def test_dup_create_same_name():
    """
    Creating two objects with the same name
    """
    ok = factory.create_object(*"primitive dup1 Dummy".split())
    assert ok
    ok = factory.create_object(*"primitive dup1 Dummy".split())
    assert not ok


def test_dup_create():
    """
    Creating property sets with unknown properties
    """
    ok = factory.create_object(*"property hana_test1: hana_attribute_1=5 hana_attribute_2=mohican".split())
    assert ok
    ok = factory.create_object(*"property hana_test2: hana_attribute_1=5s a-b-c-d=e-f-g".split())
    assert ok



#!/usr/bin/python

from __future__ import unicode_literals
from setuptools import setup

setup(name='pygments-crmsh-lexers',
      version='0.0.5',
      description='Pygments crmsh custom lexers.',
      keywords='pygments crmsh lexer',
      license='BSD',

      author='Kristoffer Gronlund',
      author_email='kgronlund@suse.com',

      url='https://github.com/ClusterLabs/crmsh',

      packages=['pygments_crmsh_lexers'],
      install_requires=['pygments>=2.0.2'],

      entry_points='''[pygments.lexers]
                      ANSIColorsLexer=pygments_crmsh_lexers:ANSIColorsLexer
                      CrmshLexer=pygments_crmsh_lexers:CrmshLexer''',

      classifiers=[
          'Environment :: Plugins',
          'Intended Audience :: Developers',
          'License :: OSI Approved :: BSD License',
          'Operating System :: OS Independent',
          'Programming Language :: Python',
          'Programming Language :: Python :: 2',
          'Programming Language :: Python :: 3',
          'Topic :: Software Development :: Libraries :: Python Modules',
      ],)



import os
import pwd
import re
import platform
import socket
import crm_script

PACKAGES = ['booth', 'corosync', 'crmsh', 'csync2', 'drbd',
            'fence-agents', 'gfs2', 'gfs2-utils', 'hawk', 'ocfs2',
            'ocfs2-tools', 'pacemaker',
            'resource-agents', 'sbd']
SERVICES = ['sshd', 'ntp', 'corosync', 'pacemaker', 'hawk']
SSH_KEY = os.path.expanduser('~/.ssh/id_rsa')
CSYNC2_KEY = '/etc/csync2/key_hagroup'
CSYNC2_CFG = '/etc/csync2/csync2.cfg'
COROSYNC_CONF = '/etc/corosync/corosync.conf'
SYSCONFIG_SBD = '/etc/sysconfig/sbd'


def rpm_info():
    'check installed packages'
    return crm_script.rpmcheck(PACKAGES)


def service_info(service):
    "Returns information about a given service"
    active, enabled = 'unknown', 'unknown'
    rc, out, err = crm_script.call(["/usr/bin/systemctl", "is-enabled", "%s.service" % (service)])
    if rc in (0, 1, 3) and out:
        enabled = out.strip()
    else:
        return {'name': service, 'error': err.strip()}
    rc, out, err = crm_script.call(["/usr/bin/systemctl", "is-active", "%s.service" % (service)])
    if rc in (0, 1, 3) and out:
        active = out.strip()
    else:
        return {'name': service, 'error': err.strip()}
    return {'name': service, 'active': active, 'enabled': enabled}


def services_info():
    'check enabled/active services'
    return [service_info(service) for service in SERVICES]

def get_user():
    return pwd.getpwuid(os.getuid()).pw_name

def sys_info():
    'system information'
    system, node, release, version, machine, processor = platform.uname()
    hostname = platform.node().split('.')[0]
    return {'system': system,
            'node': node,
            'release': release,
            'version': version,
            'machine': machine,
            'processor': processor,
            'user': get_user(),
            'hostname': hostname,
            'fqdn': socket.getfqdn()}


def net_info():
    ret = {}
    interfaces = []
    ret['interfaces'] = interfaces
    hostname = platform.node().split('.')[0]
    try:
        ip = socket.gethostbyname(hostname)
        ret['hostname'] = {'name': hostname, 'ip': ip}
    except Exception as e:
        ret['hostname'] = {'error': str(e)}
    return ret


def files_info():
    def check(fn):
        if os.path.isfile(os.path.expanduser(fn)):
            return os.path.expanduser(fn)
        return ''
    return {'ssh_key': check(SSH_KEY),
            'csync2_key': check(CSYNC2_KEY),
            'csync2_cfg': check(CSYNC2_CFG),
            'corosync_conf': check(COROSYNC_CONF),
            'sysconfig_sbd': check(SYSCONFIG_SBD),
            }


def logrotate_info():
    rc, _, _ = crm_script.call(
        'grep -r corosync.conf /etc/logrotate.d',
        shell=True)
    return {'corosync.conf': rc == 0}


def disk_info():
    rc, out, err = crm_script.call(['df'], shell=False)
    if rc == 0:
        disk_use = []
        for line in out.split('\n')[1:]:
            line = line.strip()
            if line:
                data = line.split()
                if len(data) >= 6:
                    disk_use.append((data[5], int(data[4][:-1])))
        return disk_use
    return []


def info():
    return {'rpm': rpm_info(),
            'services': services_info(),
            'system': sys_info(),
            'net': net_info(),
            'files': files_info(),
            'logrotate': logrotate_info(),
            'disk': disk_info()}


def verify(data):
    """
    Given output from info(), verifies
    as much as possible before init/add.
    """
    def check_diskspace():
        for host, info in data.items():
            for mount, percent in info['disk']:
                interesting = (mount == '/' or
                               mount.startswith('/var/log') or
                               mount.startswith('/tmp'))
                if interesting and percent > 90:
                    crm_script.exit_fail("Not enough space on %s:%s" % (host, mount))

    def check_services():
        for host, info in data.items():
            for svc in info['services']:
                if svc['name'] == 'pacemaker' and svc['active'] == 'active':
                    crm_script.exit_fail("%s already running pacemaker" % (host))
                if svc['name'] == 'corosync' and svc['active'] == 'active':
                    crm_script.exit_fail("%s already running corosync" % (host))

    def verify_host(host, info):
        if host != info['system']['hostname']:
            crm_script.exit_fail("Hostname mismatch: %s is not %s" %
                                 (host, info['system']['hostname']))

    def compare_system(systems):
        def check(value, msg):
            vals = set([system[value] for host, system in systems])
            if len(vals) > 1:
                info = ', '.join('%s: %s' % (h, system[value]) for h, system in systems)
                crm_script.exit_fail("%s: %s" % (msg, info))

        check('machine', 'Architecture differs')
        #check('release', 'Kernel release differs')
        #check('distname', 'Distribution differs')
        #check('distver', 'Distribution version differs')
        #check('version', 'Kernel version differs')

    for host, info in data.items():
        verify_host(host, info)

    compare_system((h, info['system']) for h, info in data.items())

    check_diskspace()
    check_services()


# common functions to initialize a cluster node


def is_service_enabled(name):
    info = service_info(name)
    if info.get('name') == name and info.get('enabled') == 'enabled':
        return True
    return False


def is_service_active(name):
    info = service_info(name)
    if info.get('name') == name and info.get('active') == 'active':
        return True
    return False


def install_packages(packages):
    for pkg in packages:
        try:
            crm_script.package(pkg, 'latest')
        except Exception as e:
            crm_script.exit_fail("Failed to install %s: %s" % (pkg, e))


def configure_firewall():
    _SUSE_FW_TEMPLATE = """## Name: HAE cluster ports
## Description: opens ports for HAE cluster services
TCP="%(tcp)s"
UDP="%(udp)s"
"""
    corosync_mcastport = crm_script.param('mcastport')
    if not corosync_mcastport:
        rc, out, err = crm_script.call(['crm', 'corosync', 'get', 'totem.interface.mcastport'])
        if rc == 0:
            corosync_mcastport = out.strip()

    tcp_ports = '30865 5560 7630 21064'
    udp_ports = '%s %s' % (corosync_mcastport, int(corosync_mcastport) - 1)

    # TODO: other platforms



import argparse
import dataclasses
import enum
import glob
import importlib.resources
import ipaddress
import itertools
import json
import logging
import os
import re
import shutil
import subprocess
import sys
import threading
import tempfile
import typing

import lxml.etree

from crmsh import cibquery
from crmsh import constants
from crmsh import corosync
from crmsh import corosync_config_format
from crmsh import iproute2
from crmsh import parallax
from crmsh import sh
from crmsh import utils
from crmsh import xmlutil
from crmsh.prun import prun

logger = logging.getLogger(__name__)


SAP_HANA_RESOURCE_AGENTS = {
    cibquery.ResourceAgent('ocf', 'suse', 'SAPHana'),
    cibquery.ResourceAgent('ocf', 'suse', 'SAPHanaController'),
    cibquery.ResourceAgent('ocf', 'suse', 'SAPHanaTopology'),
}


class MigrationFailure(Exception):
    pass


class CheckReturnCode(enum.IntEnum):
    ALREADY_MIGRATED = 0
    PASS_NO_AUTO_FIX = 1
    PASS_NEED_AUTO_FIX = 2
    BLOCKED_NEED_MANUAL_FIX = 3
    SSH_ERROR = 255


class CheckResultHandler:
    LEVEL_ERROR = 1
    LEVEL_WARN = 2

    def log_info(self, fmt: str, *args):
        raise NotImplementedError

    def handle_problem(self, need_auto_fix: bool, is_blocker: bool, level: int, title: str, detail: typing.Iterable[str]):
        raise NotImplementedError

    def end(self):
        raise NotImplementedError

    def to_check_return_code(self) -> CheckReturnCode:
        raise NotImplementedError


class CheckResultJsonHandler(CheckResultHandler):
    def __init__(self, indent: typing.Optional[int] = None):
        self._indent = indent
        self.json_result = {
            "problems": [],
        }

    def log_info(self, fmt: str, *args):
        logger.debug(fmt, *args)

    def handle_problem(self, need_auto_fix: bool, is_blocker: bool, level: int, title: str, detail: typing.Iterable[str]):
        self.json_result["problems"].append({
            "need_auto_fix": need_auto_fix,
            "is_blocker": is_blocker,
            "level": level,
            "title": title,
            "descriptions": detail if isinstance(detail, list) else list(detail),
        })

    def end(self):
        json.dump(
            self.json_result,
            sys.stdout,
            ensure_ascii=False,
            indent=self._indent,
        )
        sys.stdout.write('\n')

    def to_check_return_code(self) -> CheckReturnCode:
        ret = CheckReturnCode.ALREADY_MIGRATED
        for problem in self.json_result['problems']:
            if problem.get('is_blocker', False):
                ret = max(CheckReturnCode.BLOCKED_NEED_MANUAL_FIX, ret)
            elif problem.get('need_auto_fix'):
                ret = max(CheckReturnCode.PASS_NEED_AUTO_FIX, ret)
            else:
                ret = max(CheckReturnCode.PASS_NO_AUTO_FIX, ret)
        return ret


class CheckResultInteractiveHandler(CheckResultHandler):
    def __init__(self):
        self.block_migration = False
        self.has_problems = False
        self.need_auto_fix = False

    def log_info(self, fmt: str, *args):
        self.write_in_color(sys.stdout, constants.GREEN, '[INFO] ')
        print(fmt % args)

    def handle_problem(self, need_auto_fix: bool, is_blocker: bool, level:int, title: str, details: typing.Iterable[str]):
        self.has_problems = True
        self.block_migration = self.block_migration or is_blocker
        self.need_auto_fix = self.need_auto_fix or need_auto_fix
        match level:
            case self.LEVEL_ERROR:
                self.write_in_color(sys.stdout, constants.YELLOW, '[FAIL] ')
            case self.LEVEL_WARN:
                self.write_in_color(sys.stdout, constants.YELLOW, '[WARN] ')
        print(title)
        for line in details:
            sys.stdout.write('       ')
            print(line)

    @staticmethod
    def write_in_color(f, color: str, text: str):
        if f.isatty():
            f.write(color)
            f.write(text)
            f.write(constants.END)
        else:
            f.write(text)

    def end(self):
        sys.stdout.write('\n')

    def to_check_return_code(self) -> CheckReturnCode:
        if self.block_migration:
            ret = CheckReturnCode.BLOCKED_NEED_MANUAL_FIX
        elif self.need_auto_fix:
            ret = CheckReturnCode.PASS_NEED_AUTO_FIX
        elif self.has_problems:
            ret = CheckReturnCode.PASS_NO_AUTO_FIX
        else:
            ret = CheckReturnCode.ALREADY_MIGRATED
        return ret


def migrate(args: typing.Sequence[str]):
    parser = argparse.ArgumentParser(args[0])
    parser.add_argument('--local', action='store_true')
    parsed_args = parser.parse_args(args[1:])
    try:
        match _check_impl(local=parsed_args.local, json='', summary=False):
            case CheckReturnCode.ALREADY_MIGRATED:
                logger.info("This cluster works on SLES 16. No migration is needed.")
                return 0
            case CheckReturnCode.PASS_NO_AUTO_FIX:
                logger.info("This cluster works on SLES 16 with some warnings. Please fix the remaining warnings manually.")
                return 0
            case CheckReturnCode.PASS_NEED_AUTO_FIX:
                logger.info('Starting migration...')
                migrate_corosync_conf(local=parsed_args.local)
                logger.info('Finished migration.')
                return 0
            case _:
                raise MigrationFailure('Unable to start migration.')
    except MigrationFailure as e:
        logger.error('%s', e)
        return 1


def check(args: typing.Sequence[str]) -> int:
    parser = argparse.ArgumentParser(args[0])
    parser.add_argument('--json', nargs='?', const='pretty', choices=['oneline', 'pretty'])
    parser.add_argument('--local', action='store_true')
    parsed_args = parser.parse_args(args[1:])
    ret = _check_impl(parsed_args.local or parsed_args.json is not None, parsed_args.json, parsed_args.json is None)
    if not parsed_args.json:
        print('****** summary ******')
        match ret:
            case CheckReturnCode.ALREADY_MIGRATED:
                CheckResultInteractiveHandler.write_in_color(sys.stdout, constants.GREEN, '[INFO]')
                sys.stdout.write(' This cluster works on SLES 16. No migration is needed.\n')
            case CheckReturnCode.PASS_NO_AUTO_FIX:
                CheckResultInteractiveHandler.write_in_color(sys.stdout, constants.GREEN, '[PASS]')
                sys.stdout.write(' This cluster works on SLES 16 with some warnings. Please fix the remaining warnings manually.\n')
            case CheckReturnCode.PASS_NEED_AUTO_FIX:
                CheckResultInteractiveHandler.write_in_color(sys.stdout, constants.GREEN, '[INFO]')
                sys.stdout.write(' Please run "crm cluster health sles16 --fix" on on any one of above nodes.\n')
                CheckResultInteractiveHandler.write_in_color(sys.stdout, constants.GREEN, '[PASS]')
                sys.stdout.write(' This cluster is good to migrate to SLES 16.\n')
            case _:
                CheckResultInteractiveHandler.write_in_color(sys.stdout, constants.RED, '[FAIL]')
                sys.stdout.write(' The pacemaker cluster stack can not migrate to SLES 16.\n')
    return ret


def _check_impl(local: bool, json: str, summary: bool) -> CheckReturnCode:
    assert not summary or not bool(json)
    assert local or not bool(json)
    if not local:
        check_remote_yield = check_remote()
        next(check_remote_yield)
    else:
        check_remote_yield = itertools.repeat(0)
    match json:
        case 'oneline':
            handler = CheckResultJsonHandler()
        case 'pretty':
            handler = CheckResultJsonHandler(indent=2)
        case _:
            handler = CheckResultInteractiveHandler()
    if local:
        check_remote_yield = itertools.repeat(0)
        check_local(handler)
    else:
        check_remote_yield = check_remote()
        next(check_remote_yield)
        print('------ node: localhost ------')
        check_local(handler)
        print('\n------ cib ------')
        check_global(handler)
    handler.end()
    ret = handler.to_check_return_code()
    if check_remote_yield:
        remote_ret = next(check_remote_yield)
        ret = max(remote_ret, ret)
    return ret


def check_local(handler: CheckResultHandler):
    check_dependency_version(handler)
    check_unsupported_corosync_features(handler)


def check_remote():
    handler = CheckResultInteractiveHandler()
    class CheckRemoteThread(threading.Thread):
        def run(self):
            self.result = prun.prun({
                node: 'crm cluster health sles16 --local --json=oneline'
                for node in utils.list_cluster_nodes_except_me()
            })
    prun_thread = CheckRemoteThread()
    prun_thread.start()
    yield
    prun_thread.join()
    ret = CheckReturnCode.ALREADY_MIGRATED
    for host, result in prun_thread.result.items():
        sys.stdout.write(f'------ node: {host} ------\n')
        match result:
            case prun.SSHError() as e:
                handler.write_in_color(
                    sys.stdout, constants.YELLOW,
                    str(e)
                )
                sys.stdout.write('\n')
                ret = CheckReturnCode.SSH_ERROR
            case prun.ProcessResult() as result:
                try:
                    check_result = json.loads(result.stdout.decode('utf-8'))
                except (UnicodeDecodeError, json.JSONDecodeError):
                    print(result.stdout.decode('utf-8', 'backslashreplace'))
                    handler.write_in_color(
                        sys.stderr, constants.YELLOW,
                        result.stderr.decode('utf-8', 'backslashreplace')
                    )
                    sys.stdout.write('\n')
                    # cannot pass the exit status through,
                    # as all failed exit status become 1 in ui_context.Context.run()
                    ret = CheckReturnCode.BLOCKED_NEED_MANUAL_FIX
                else:
                    handler = CheckResultInteractiveHandler()
                    problems = check_result.get("problems", list())
                    for problem in problems:
                        handler.handle_problem(
                            problem.get("need_auto_fix", False), problem.get("is_blocker", False),
                            problem.get("level", handler.LEVEL_ERROR),
                            problem.get("title", ""), problem.get("descriptions"),
                        )
                    handler.end()
                    ret = handler.to_check_return_code()
    yield ret


def check_global(handler: CheckResultHandler):
    cib = xmlutil.text2elem(sh.LocalShell().get_stdout_or_raise_error(None, 'crm configure show xml'))
    check_cib_schema_version(handler, cib)
    check_unsupported_resource_agents(handler, cib)


def check_dependency_version(handler: CheckResultHandler):
    handler.log_info('Checking dependency version...')
    shell = sh.LocalShell()
    out = shell.get_stdout_or_raise_error(None, 'corosync -v')
    _check_version_range(
        handler,
        'Corosync', (3,),
        re.compile(r"version\s+'(\d+(?:\.\d+)*)'"),
        shell.get_stdout_or_raise_error(None, 'corosync -v'),
    )


def _check_version_range(
        handler: CheckResultHandler, component_name: str,
        minimum: tuple,
        pattern,
        text: str,
):
    match = pattern.search(text)
    if not match:
        handler.handle_problem(
            True, True, handler.LEVEL_ERROR,
            f'{component_name} version not supported',
            [
                'Unknown version:',
                text,
            ],
        )
    else:
        version = tuple(int(x) for x in match.group(1).split('.'))
        if not minimum <= version:
            handler.handle_problem(
                True, True,  handler.LEVEL_ERROR,
                f'{component_name} version not supported', [
                    'Supported version: {} >= {}'.format(
                        component_name,
                        '.'.join(str(x) for x in minimum),
                    ),
                    f'Actual version:    {component_name} == {match.group(1)}',
                ],
            )


def check_unsupported_corosync_features(handler: CheckResultHandler):
    handler.log_info("Checking used corosync features...")
    conf_path = corosync.conf()
    with open(conf_path, 'r', encoding='utf-8') as f:
        config = corosync_config_format.DomParser(f).dom()
        corosync.ConfParser.transform_dom_with_list_schema(config)
    if config['totem'].get('rrp_mode', None) in {'active', 'passive'}:
        handler.handle_problem(
            True, False, handler.LEVEL_WARN,
            'Corosync RRP is deprecated in corosync 3.', [
                'Run "crm health sles16 --fix" to migrate it to knet multilink.',
            ],
        )
    _check_unsupported_corosync_transport(handler, config)


def _check_unsupported_corosync_transport(handler: CheckResultHandler, dom):
    transport = dom['totem'].get('transport', None)
    if transport == 'knet':
        return
    if transport is None:
        try:
            dom['totem']['interface'][0]['bindnetaddr']
        except KeyError:
            # looks like a corosync 3 config
            return
    handler.handle_problem(
        True, False, handler.LEVEL_WARN,
        f'Corosync transport "{transport}" is deprecated in corosync 3. Please use knet.', [],
    )


def migrate_corosync_conf(local: bool):
    conf_path = corosync.conf()
    with open(conf_path, 'r', encoding='utf-8') as f:
        config = corosync_config_format.DomParser(f).dom()
        corosync.ConfParser.transform_dom_with_list_schema(config)
    logger.info('Migrating corosync configuration...')
    migrate_corosync_conf_impl(config)
    shutil.copy(conf_path, conf_path + '.bak')
    with utils.open_atomic(conf_path, 'w', fsync=True, encoding='utf-8') as f:
        corosync_config_format.DomSerializer(config, f)
        os.fchmod(f.fileno(), 0o644)
    logger.info(
        'Finish migrating corosync configuration. The original configuration is renamed to %s.bak',
        os.path.basename(conf_path),
    )
    if not local:
        for host, result in prun.pcopy_to_remote(
                conf_path,
                utils.list_cluster_nodes_except_me(), conf_path,
                atomic_write=True,
        ).items():
            match result:
                case None:
                    pass
                case prun.PRunError() as e:
                    logger.error("Failed to copy crmsh.conf to host %s: %s", host, e)


def migrate_corosync_conf_impl(config):
    assert 'totem' in config
    corosync.ConfParser.transform_dom_with_list_schema(config)
    migrate_transport(config)
    migrate_crypto(config)
    migrate_rrp(config)
    # TODO: other migrations


def migrate_transport(dom):
    match dom['totem'].get('transport', None):
        case 'knet':
            return
        case 'udpu':
            migrate_udpu(dom)
        case 'udp':
            migrate_multicast(dom)
        case _:
            # corosync 2 defaults to "udp"
            try:
                dom['totem']['interface'][0]['bindnetaddr']
            except KeyError:
                # looks like a corosync 3 config
                pass
            if 'nodelist' not in dom:
                migrate_multicast(dom)
            else:
                # looks like a corosync 3 config
                pass


def migrate_udpu(dom):
    dom['totem']['transport'] = 'knet'
    if 'interface' in dom['totem']:
        for interface in dom['totem']['interface']:
            _migrate_totem_interface(interface)
    if 'quorum' in dom:
        dom['quorum'].pop('expected_votes', None)
    logger.info("Upgrade totem.transport to knet.")


def migrate_multicast(dom):
    dom['totem']['transport'] = 'knet'
    for interface in dom['totem']['interface']:
        _migrate_totem_interface(interface)
    logger.info("Generating nodelist according to CIB...")
    with open(constants.CIB_RAW_FILE, 'rb') as f:
        cib = lxml.etree.parse(f)
    cib_nodes = cibquery.get_cluster_nodes(cib)
    assert 'nodelist' not in dom
    nodelist = list()
    node_interfaces = {
        x[0]: iproute2.IPAddr(json.loads(x[1][1]))
        for x in parallax.parallax_call([x.uname for x in cib_nodes], 'ip -j addr')
        if x[1][0] == 0
    }
    with tempfile.TemporaryDirectory(prefix='crmsh-migration-') as dir_name:
        node_configs = {
            x[0]: x[1]
            for x in parallax.parallax_slurp([x.uname for x in cib_nodes], dir_name, corosync.conf())
        }
        for node in cib_nodes:
            assert node.uname in node_configs
            bindnetaddr_fixer = _BindnetaddrFixer(node_interfaces[node.uname].interfaces())
            with open(node_configs[node.uname], 'r', encoding='utf-8') as f:
                root = corosync_config_format.DomParser(f).dom()
                corosync.ConfParser.transform_dom_with_list_schema(root)
                interfaces = root['totem']['interface']
                addresses = {f'ring{i}_addr': bindnetaddr_fixer.fix_bindnetaddr(x['bindnetaddr']) for i, x in enumerate(interfaces)}
                logger.info("Node %s: %s: %s", node.node_id, node.uname, addresses)
                nodelist.append({
                    'nodeid': node.node_id,
                    'name': node.uname,
                } | addresses)
    dom['nodelist'] = {'node': nodelist}
    if 'quorum' in dom:
        dom['quorum'].pop('expected_votes', None)
        logger.info("Unset quorum.expected_votes.")
    logger.info("Upgrade totem.transport to knet.")


def _migrate_totem_interface(interface):
    # remove udp-only items
    interface.pop('mcastaddr', None)
    interface.pop('bindnetaddr', None)
    interface.pop('broadcast', None)
    interface.pop('ttl', None)
    ringnumber = interface.pop('ringnumber', None)
    if ringnumber is not None:
        interface['linknumber'] = ringnumber


class _BindnetaddrFixer:
    # crmsh generates incorrect bindnetaddr when joining a corosync 2 multicast cluster
    def __init__(self, interfaces: typing.Iterable[iproute2.IPInterface]):
        self._interface_addresses = {addr_info for interface in interfaces for addr_info in interface.addr_info}

    def fix_bindnetaddr(self, bindnetaddr: str):
        bind_address = ipaddress.ip_address(bindnetaddr)
        for interface_address in self._interface_addresses:
            if bind_address in interface_address.network:
                return str(interface_address.ip)
        return bindnetaddr


def migrate_crypto(dom):
    try:
        # corosync 3 change the default hash algorithm to sha256 when `secauth` is enabled
        if dom['totem'].get('crypto_hash', None) == 'sha1':
            dom['totem']['crypto_hash'] = 'sha256'
            logger.info('Upgrade totem.crypto_hash from "sha1" to "sha256".')
    except KeyError:
        dom['totem']['crypto_hash'] = 'sha256'


def migrate_rrp(dom):
    try:
        nodes = dom['nodelist']['node']
    except KeyError:
        return
    is_rrp = any('ring1_addr' in node for node in nodes)
    if not is_rrp:
        return
    try:
        rrp_mode = dom['totem']['rrp_mode']
        del dom['totem']['rrp_mode']
        if rrp_mode == 'active':
            dom['totem']['link_mode'] = 'active'
    except KeyError:
        pass
    assert all('nodeid' in node for node in nodes)
    if any('name' not in node for node in nodes):
        populate_node_name(nodes)


def populate_node_name(nodelist):
    # cannot use utils.list_cluster_nodes, as pacemaker is not running
    with open(constants.CIB_RAW_FILE, 'rb') as f:
        cib = lxml.etree.parse(f)
    cib_nodes = {node.node_id: node for node in cibquery.get_cluster_nodes(cib)}
    for node in nodelist:
        node_id = int(node['nodeid'])
        node['name'] = cib_nodes[node_id].uname


def check_unsupported_resource_agents(handler: CheckResultHandler, cib: lxml.etree.Element):
    handler.log_info("Checking used resource agents...")
    ocf_resource_agents = list()
    stonith_resource_agents = list()
    class_unsupported_resource_agents = list()
    for resource_agent in cibquery.get_configured_resource_agents(cib):
        if resource_agent.m_class == 'ocf':
            ocf_resource_agents.append(resource_agent)
        elif resource_agent.m_class == 'stonith':
            stonith_resource_agents.append(resource_agent)
        elif resource_agent.m_class in {'lsb', 'service'}:
            class_unsupported_resource_agents.append(resource_agent)
        else:
            logger.debug('Unrecognized resource agent class: %s', resource_agent)
    unsupported_resource_agents = UnsupportedResourceAgentDetector()
    _check_saphana_resource_agent(handler, ocf_resource_agents)
    _check_removed_resource_agents(
        handler,
        "resource agents",
        unsupported_resource_agents,
        (agent for agent in ocf_resource_agents if agent not in SAP_HANA_RESOURCE_AGENTS),
    )
    _check_removed_resource_agents(
        handler,
        "fence agents",
        unsupported_resource_agents,
        stonith_resource_agents,
    )
    if class_unsupported_resource_agents:
        handler.handle_problem(
            False, True, handler.LEVEL_ERROR,
            'The following resource agents from class "lsb" or "service" are not supported in SLES 16.',
            ('* ' + ':'.join(x for x in dataclasses.astuple(resource_agent) if x is not None) for resource_agent in class_unsupported_resource_agents)
        )
    _check_ocfs2(handler, cib)


def _check_saphana_resource_agent(handler: CheckResultHandler, resource_agents: typing.Iterable[cibquery.ResourceAgent]):
    # "SAPHana" appears only in SAPHanaSR Classic
    has_sap_hana_sr_resources = any(agent in SAP_HANA_RESOURCE_AGENTS for agent in resource_agents)
    if has_sap_hana_sr_resources:
        if 0 != subprocess.run(
            ['rpm', '-q', 'SAPHanaSR-angi'],
            stdin=subprocess.DEVNULL,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
        ).returncode:
            handler.handle_problem(
                False, True, handler.LEVEL_ERROR,
                'SAPHanaSR Classic is removed in SLES 16.', [
                    'Before migrating to SLES 16, replace it with SAPHanaSR-angi.',
                ],
            )


class UnsupportedResourceAgentDetector:
    @dataclasses.dataclass(frozen=True)
    class UnsupportedState:
        alternative: typing.Optional[cibquery.ResourceAgent]
        is_deprecated: bool

    def __init__(self):
        self._unsupported = dict()
        with importlib.resources.files('crmsh').joinpath('migration-unsupported-resource-agents.txt').open(
                'r', encoding='ascii',
        ) as r:
            for line in r:
                parts = line.strip().split(',', 3)
                parts.extend(itertools.repeat('', 2))
                unsupported = self.__resource_agent_from_str(parts[0])
                alternative = parts[1]
                is_deprecated = parts[2] == "deprecated"
                self._unsupported[unsupported] = self.UnsupportedState(
                    self.__resource_agent_from_str(alternative) if alternative != '' else None,
                    is_deprecated,
                )

    def get_unsupported_state(self, resource_agent):
        return self._unsupported.get(resource_agent)

    @staticmethod
    def __resource_agent_from_str(s: str):
        parts = s.split(':', 3)
        m_class = parts[0]
        m_provider = parts[1] if len(parts) == 3 else None
        m_type = parts[-1]
        return cibquery.ResourceAgent(m_class, m_provider, m_type)


def _check_removed_resource_agents(
        handler: CheckResultHandler,
        agent_type_message: str,
        unsupported_resource_agents: UnsupportedResourceAgentDetector,
        resource_agents: typing.Iterable[cibquery.ResourceAgent],
):
    unsupported: typing.List[
        typing.Tuple[cibquery.ResourceAgent, UnsupportedResourceAgentDetector.UnsupportedState]] = list()
    deprecated: typing.List[
        typing.Tuple[cibquery.ResourceAgent, UnsupportedResourceAgentDetector.UnsupportedState]] = list()
    for x in resource_agents:
        unsupported_state = unsupported_resource_agents.get_unsupported_state(x)
        if unsupported_state is None:
            pass
        elif unsupported_state.is_deprecated:
            deprecated.append((x, unsupported_state))
        else:
            unsupported.append((x, unsupported_state))
    if unsupported:
        handler.handle_problem(
            False, True, handler.LEVEL_ERROR,
            f'The following {agent_type_message} are removed in SLES 16.', [
                '* {}{}'.format(
                    ':'.join(x for x in dataclasses.astuple(resource_agent) if x is not None),
                    ': please replace it with {}'.format(
                        ':'.join(x for x in dataclasses.astuple(unsupported_state.alternative) if x is not None)
                ) if unsupported_state.alternative is not None else ''
            )
            for resource_agent, unsupported_state in unsupported
        ])
    if deprecated:
        handler.handle_problem(
            False, False, handler.LEVEL_WARN,
            f'The following {agent_type_message} are deprecated in SLES 16.', [
            '* {}{}'.format(
                ':'.join(x for x in dataclasses.astuple(resource_agent) if x is not None),
                ': please replace it with {}'.format(
                    ':'.join(x for x in dataclasses.astuple(unsupported_state.alternative) if x is not None)
                ) if unsupported_state.alternative is not None else ''
            )
            for resource_agent, unsupported_state in deprecated
        ])


def _check_ocfs2(handler: CheckResultHandler, cib: lxml.etree.Element):
    if cibquery.has_primitive_filesystem_with_fstype(cib, 'ocfs2'):
       handler.handle_problem(
           False, True, handler.LEVEL_ERROR,
           'OCFS2 is not supported in SLES 16. Please use GFS2.', [],
       )

def check_cib_schema_version(handler: CheckResultHandler, cib: lxml.etree.Element):
    schema_version = cib.get('validate-with')
    if schema_version is None:
        handler.handle_problem(
            False, False, handler.LEVEL_WARN,
            "The CIB is validated with unknown schema version.", []
        )
        return
    version_match = re.match(r'^pacemaker-(\d+)\.(\d+)$', schema_version)
    if version_match is None:
        handler.handle_problem(
            False, False, handler.LEVEL_WARN,
            f"The CIB is validated with unknown schema version {schema_version}", []
        )
        return
    version = tuple(int(x) for x in version_match.groups())
    latest_schema_version = _get_latest_cib_schema_version()
    if version != latest_schema_version:
        handler.handle_problem(
            False, False, handler.LEVEL_WARN,
            "The CIB is not validated with the latest schema version.", [
                f'* Latest version:  {".".join(str(i) for i in latest_schema_version)}',
                f'* Current version: {".".join(str(i) for i in version)}',
                'Please run "crm configure upgrade force" to upgrade to the latest version.',
            ]
        )


def _get_latest_cib_schema_version() -> tuple[int, int]:
    return max(tuple(int(s) for s in x.groups()) for x in (
        re.match(r'^pacemaker-(\d+)\.(\d+)\.rng$', filename)
        for filename in glob.iglob('pacemaker-*.rng', root_dir='/usr/share/pacemaker')
    ) if x is not None)
